# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import pyuavcan.transport


SessionFinalizer = typing.Callable[[], None]


_logger = logging.getLogger(__name__)


class CANSession:
    def __init__(self, finalizer: SessionFinalizer):
        self._close_finalizer: typing.Optional[SessionFinalizer] = finalizer

    def _raise_if_closed(self) -> None:
        if self._close_finalizer is None:
            raise pyuavcan.transport.ResourceClosedError(
                f"The requested action cannot be performed because the session object {self} is closed"
            )

    def close(self) -> None:
        fin = self._close_finalizer
        if fin is not None:
            self._close_finalizer = None
            fin()



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import pyuavcan.transport


class PortClosedError(pyuavcan.transport.ResourceClosedError):
    """
    Raised when an attempt is made to use a presentation-layer session instance that has been closed.
    Observe that it is a specialization of the corresponding transport-layer error type.
    Double-close is NOT an error, so closing the same instance twice will not result in this exception being raised.
    """


class RequestTransferIDVariabilityExhaustedError(pyuavcan.transport.TransportError):
    """
    Raised when an attempt is made to invoke more concurrent requests that supported by the transport layer.
    For CAN, the number is 32; for some transports the number is unlimited (technically, there is always a limit,
    but for some transports, such as the serial transport, it is unreachable in practice).
    """



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import random
from typing import Optional, Union
from pathlib import Path
import logging
import pyuavcan
from ._node import Node, NodeInfo
from . import register
from ._transport_factory import make_transport
from ._registry_factory import make_registry


class MissingTransportConfigurationError(register.MissingRegisterError):
    pass


class SimpleNode(Node):
    def __init__(
        self,
        presentation: pyuavcan.presentation.Presentation,
        info: NodeInfo,
        registry: register.Registry,
    ) -> None:
        self._presentation = presentation
        self._info = info
        self._registry = registry
        super().__init__()

    @property
    def presentation(self) -> pyuavcan.presentation.Presentation:
        return self._presentation

    @property
    def info(self) -> NodeInfo:
        return self._info

    @property
    def registry(self) -> register.Registry:
        return self._registry


def make_node(
    info: NodeInfo,
    registry: Union[None, register.Registry, str, Path] = None,
    *,
    transport: Optional[pyuavcan.transport.Transport] = None,
    reconfigurable_transport: bool = False,
) -> Node:
    """
    Initialize a new node by parsing the configuration encoded in the UAVCAN registers.

    Aside from the registers that encode the transport configuration (which are documented in :func:`make_transport`),
    the following registers are considered (if they don't exist, they are automatically created).
    They are split into groups by application-layer function they configure.

    ..  list-table:: General
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.node.unique_id``
          - ``unstructured``
          - The unique-ID of the local node.
            This register is only used if the caller did not set ``unique_id`` in ``info``.
            If not defined, a new random value is generated and stored as immutable
            (therefore, if no persistent register file is used, a new unique-ID is generated at every launch, which
            may be undesirable in some applications, particularly those that require PnP node-ID allocation).

        * - ``uavcan.node.description``
          - ``string``
          - As defined by the UAVCAN Specification, this standard register is intended to store a human-friendly
            description of the node.
            Empty by default and never accessed by the library, since it is intended mostly for remote use.

    ..  list-table:: :mod:`pyuavcan.application.diagnostic`
        :widths: 1 1 9
        :header-rows: 1

        * - Register name
          - Register type
          - Register semantics

        * - ``uavcan.diagnostic.severity``
          - ``natural8[1]``
          - If the value is a valid severity level as defined in ``uavcan.diagnostic.Severity``,
            the node will publish its application log records of matching severity level to the standard subject
            ``uavcan.diagnostic.Record`` using :class:`pyuavcan.application.diagnostic.DiagnosticPublisher`.
            This is done by installing a root handler in :mod:`logging`.
            Disabled by default.

        * - ``uavcan.diagnostic.timestamp``
          - ``bit[1]``
          - If true, the published log messages will initialize the synchronized ``timestamp`` field
            from the log record timestamp provided by the :mod:`logging` library.
            This is only safe if the UAVCAN network is known to be synchronized on the same time system as the
            wall clock of the local computer.
            Otherwise, the timestamp is left at zero (which means "unknown" per Specification).
            Disabled by default.

    Additional application-layer functions and their respective registers may be added later.

    :param info:
        Response object to ``uavcan.node.GetInfo``. The following fields will be populated automatically:

        - ``protocol_version`` from :data:`pyuavcan.UAVCAN_SPECIFICATION_VERSION`.

        - If not set by the caller: ``unique_id`` is read from register as specified above.

        - If not set by the caller: ``name`` is constructed from hex-encoded unique-ID like:
          ``anonymous.b0228a49c25ff23a3c39915f81294622``.

    :param registry:
        If this is an instance of :class:`pyuavcan.application.register.Registry`, it is used as-is
        (ownership is taken).
        Otherwise, this is a register file path (or None) that is passed over to
        :func:`pyuavcan.application.make_registry`
        to construct the registry instance for this node.
        This instance will be available under :class:`pyuavcan.application.Node.registry`.

    :param transport:
        If not provided (default), a new transport instance will be initialized based on the available registers using
        :func:`make_transport`.
        If provided, the node will be constructed with this transport instance and take its ownership.
        In the latter case, existence of transport-related registers will NOT be ensured.

    :param reconfigurable_transport:
        If True, the node will be constructed with :mod:`pyuavcan.transport.redundant`,
        which permits runtime reconfiguration.
        If the transport argument is given and it is not a redundant transport, it will be wrapped into one.
        Also see :func:`make_transport`.

    :raises:
        - :class:`pyuavcan.application.register.MissingRegisterError` if a register is expected but cannot be found,
          or if no transport is configured.
        - :class:`pyuavcan.application.register.ValueConversionError` if a register is found but its value
          cannot be converted to the correct type, or if the value of an environment variable for a register
          is invalid or incompatible with the register's type
          (e.g., an environment variable set to ``Hello world`` cannot initialize a register  of type ``real64[3]``).
        - Also see :func:`make_transport`.

    ..  note::

        Consider extending this factory with a capability to automatically run the node-ID allocation client
        :class:`pyuavcan.application.plug_and_play.Allocatee` if ``uavcan.node.id`` is not set.

        Until this is implemented, to run the allocator one needs to construct the transport manually using
        :func:`make_transport` and :func:`make_registry`,
        then run the allocation client, then invoke this factory again with the above-obtained Registry instance,
        having done ``registry["uavcan.node.id"] = allocated_node_id`` beforehand.

        While tedious, this is not that much of a problem because the PnP protocol is mostly intended for
        hardware nodes rather than software ones.
        A typical software node would normally receive its node-ID at startup (see also Yakut Orchestrator).
    """
    from pyuavcan.transport.redundant import RedundantTransport

    if not isinstance(registry, register.Registry):
        registry = make_registry(registry)
    assert isinstance(registry, register.Registry)

    def init_transport() -> pyuavcan.transport.Transport:
        assert isinstance(registry, register.Registry)
        if transport is None:
            out = make_transport(registry, reconfigurable=reconfigurable_transport)
            if out is not None:
                return out
            raise MissingTransportConfigurationError(
                "Available registers do not encode a valid transport configuration"
            )
        if not isinstance(transport, RedundantTransport) and reconfigurable_transport:
            out = RedundantTransport()
            out.attach_inferior(transport)
            return out
        return transport

    # Populate certain fields of the node info structure automatically and create standard registers.
    info.protocol_version.major, info.protocol_version.minor = pyuavcan.UAVCAN_SPECIFICATION_VERSION
    if info.unique_id.sum() == 0:
        info.unique_id = bytes(
            registry.setdefault(
                "uavcan.node.unique_id",
                register.Value(unstructured=register.Unstructured(random.getrandbits(128).to_bytes(16, sys.byteorder))),
            )
        )
    registry.setdefault("uavcan.node.description", register.Value(string=register.String()))

    if len(info.name) == 0:
        info.name = "anonymous." + info.unique_id.tobytes().hex()

    # Construct the node and its application-layer functions.
    node = SimpleNode(pyuavcan.presentation.Presentation(init_transport()), info, registry)
    _make_diagnostic_publisher(node)

    return node


def _make_diagnostic_publisher(node: Node) -> None:
    from .diagnostic import DiagnosticSubscriber, DiagnosticPublisher

    uavcan_severity = int(
        node.registry.setdefault("uavcan.diagnostic.severity", register.Value(natural8=register.Natural8([0xFF])))
    )
    timestamping_enabled = bool(
        node.registry.setdefault("uavcan.diagnostic.timestamp", register.Value(bit=register.Bit([False])))
    )

    try:
        level = DiagnosticSubscriber.SEVERITY_UAVCAN_TO_PYTHON[uavcan_severity]
    except LookupError:
        return

    diag_publisher = DiagnosticPublisher(node, level=level)
    diag_publisher.timestamping_enabled = timestamping_enabled

    logging.root.addHandler(diag_publisher)
    node.add_lifetime_hooks(None, lambda: logging.root.removeHandler(diag_publisher))


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import sys
import copy
import socket as socket_
import typing
import asyncio
import logging
import pyuavcan
from pyuavcan.transport import Timestamp, ServiceDataSpecifier
from .._frame import UDPFrame


_IGNORE_OS_ERROR_ON_SEND = sys.platform.startswith("win")
r"""
On Windows, multicast output sockets have a weird corner case.
If the output interface is set to the loopback adapter and there are no registered listeners for the specified
multicast group, an attempt to send data to that group will fail with a "network unreachable" error.
Here is an example::

    import socket, asyncio
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.bind(('127.1.2.3', 0))
    s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton('127.1.2.3'))
    s.sendto(b'\xaa\xbb\xcc', ('127.5.5.5', 1234))          # Success
    s.sendto(b'\xaa\xbb\xcc', ('239.1.2.3', 1234))          # OSError
    # OSError: [WinError 10051] A socket operation was attempted to an unreachable network
    loop = asyncio.get_event_loop()
    loop.run_until_complete(loop.sock_sendall(s, b'abc'))   # OSError
    # OSError: [WinError 1231] The network location cannot be reached
"""

_logger = logging.getLogger(__name__)


class UDPFeedback(pyuavcan.transport.Feedback):
    def __init__(self, original_transfer_timestamp: Timestamp, first_frame_transmission_timestamp: Timestamp):
        self._original_transfer_timestamp = original_transfer_timestamp
        self._first_frame_transmission_timestamp = first_frame_transmission_timestamp

    @property
    def original_transfer_timestamp(self) -> Timestamp:
        return self._original_transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> Timestamp:
        return self._first_frame_transmission_timestamp


class UDPOutputSession(pyuavcan.transport.OutputSession):
    """
    The output session logic is extremely simple because most of the work is handled by the UDP/IP
    stack of the operating system.
    Here we just split the transfer into frames, encode the frames, and write them into the socket one by one.
    If the transfer multiplier is greater than one (for unreliable networks),
    we repeat that the required number of times.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        mtu: int,
        multiplier: int,
        sock: socket_.socket,
        loop: asyncio.AbstractEventLoop,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly. Instead, use the factory method.
        Instances take ownership of the socket.
        """
        self._closed = False
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._mtu = int(mtu)
        self._multiplier = int(multiplier)
        self._sock = sock
        self._loop = loop
        self._finalizer = finalizer
        self._feedback_handler: typing.Optional[typing.Callable[[pyuavcan.transport.Feedback], None]] = None
        self._statistics = pyuavcan.transport.SessionStatistics()
        if self._multiplier < 1:  # pragma: no cover
            raise ValueError(f"Invalid transfer multiplier: {self._multiplier}")

        assert (
            specifier.remote_node_id is None
            if isinstance(specifier.data_specifier, pyuavcan.transport.MessageDataSpecifier)
            else True
        ), "Internal protocol violation: cannot unicast a message transfer"
        assert (
            specifier.remote_node_id is not None if isinstance(specifier.data_specifier, ServiceDataSpecifier) else True
        ), "Internal protocol violation: cannot broadcast a service transfer"

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        if self._closed:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

        def construct_frame(index: int, end_of_transfer: bool, payload: memoryview) -> UDPFrame:
            return UDPFrame(
                priority=transfer.priority,
                transfer_id=transfer.transfer_id,
                index=index,
                end_of_transfer=end_of_transfer,
                payload=payload,
            )

        frames = [
            fr.compile_header_and_payload()
            for fr in pyuavcan.transport.commons.high_overhead_transport.serialize_transfer(
                transfer.fragmented_payload, self._mtu, construct_frame
            )
        ]
        _logger.debug("%s: Sending transfer: %s; current stats: %s", self, transfer, self._statistics)
        tx_timestamp = await self._emit(frames, monotonic_deadline)
        if tx_timestamp is None:
            return False

        self._statistics.transfers += 1

        # Once we have transmitted at least one copy of a multiplied transfer, it's a success.
        # We don't care if redundant copies fail.
        for _ in range(self._multiplier - 1):
            if not await self._emit(frames, monotonic_deadline):
                break

        if self._feedback_handler is not None:
            try:
                self._feedback_handler(
                    UDPFeedback(
                        original_transfer_timestamp=transfer.timestamp, first_frame_transmission_timestamp=tx_timestamp
                    )
                )
            except Exception as ex:  # pragma: no cover
                _logger.exception(
                    "Unhandled exception in the output session feedback handler %s: %s", self._feedback_handler, ex
                )

        return True

    def enable_feedback(self, handler: typing.Callable[[pyuavcan.transport.Feedback], None]) -> None:
        self._feedback_handler = handler

    def disable_feedback(self) -> None:
        self._feedback_handler = None

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> pyuavcan.transport.SessionStatistics:
        return copy.copy(self._statistics)

    def close(self) -> None:
        if not self._closed:
            self._closed = True
            try:
                self._sock.close()
            finally:
                self._finalizer()

    @property
    def socket(self) -> socket_.socket:
        """
        Provides access to the underlying UDP socket.
        """
        return self._sock

    async def _emit(
        self, header_payload_pairs: typing.Sequence[typing.Tuple[memoryview, memoryview]], monotonic_deadline: float
    ) -> typing.Optional[Timestamp]:
        """
        Returns the transmission timestamp of the first frame (which is the transfer timestamp) on success.
        Returns None if at least one frame could not be transmitted.
        """
        ts: typing.Optional[Timestamp] = None
        for index, (header, payload) in enumerate(header_payload_pairs):
            try:
                # TODO: concatenation is inefficient. Use vectorized IO via sendmsg() instead!
                await asyncio.wait_for(
                    self._loop.sock_sendall(self._sock, b"".join((header, payload))),
                    timeout=monotonic_deadline - self._loop.time(),
                )

                # TODO: use socket timestamping when running on Linux (Windows does not support timestamping).
                # Depending on the chosen approach, timestamping on Linux may require us to launch a new thread
                # reading from the socket's error message queue and then matching the returned frames with a
                # pending loopback registry, kind of like it's done with CAN.
                ts = ts or Timestamp.now()

            except (asyncio.TimeoutError, asyncio.CancelledError):
                self._statistics.drops += len(header_payload_pairs) - index
                return None
            except Exception as ex:
                if _IGNORE_OS_ERROR_ON_SEND and isinstance(ex, OSError) and self._sock.fileno() >= 0:
                    # Windows compatibility workaround -- if there are no registered multicast receivers on the
                    # loopback interface, send() may raise WinError 1231 or 10051. This error shall be suppressed.
                    _logger.debug(
                        "%r: Socket send error ignored (the likely cause is that there are no known receivers "
                        "on the other end of the link): %r",
                        self,
                        ex,
                    )
                    # To suppress the error properly, we have to pretend that the data was actually transmitted,
                    # so we populate the timestamp with a phony value anyway.
                    ts = ts or Timestamp.now()
                else:
                    self._statistics.errors += 1
                    raise

            self._statistics.frames += 1
            self._statistics.payload_bytes += len(payload)

        return ts


def _unittest_output_session() -> None:
    from pytest import raises
    from pyuavcan.transport import OutputSessionSpecifier, MessageDataSpecifier, Priority
    from pyuavcan.transport import PayloadMetadata, SessionStatistics, Feedback, Transfer

    ts = Timestamp.now()
    loop = asyncio.get_event_loop()
    run_until_complete = loop.run_until_complete
    finalized = False

    def do_finalize() -> None:
        nonlocal finalized
        finalized = True

    def check_timestamp(t: Timestamp) -> bool:
        now = Timestamp.now()
        s = ts.system_ns <= t.system_ns <= now.system_ns
        m = ts.monotonic_ns <= t.monotonic_ns <= now.system_ns
        return s and m

    destination_endpoint = "127.100.0.1", 25406

    sock_rx = socket_.socket(socket_.AF_INET, socket_.SOCK_DGRAM)
    sock_rx.bind(destination_endpoint)
    sock_rx.settimeout(1.0)

    def make_sock() -> socket_.socket:
        sock = socket_.socket(socket_.AF_INET, socket_.SOCK_DGRAM)
        sock.bind(("127.100.0.2", 0))
        sock.connect(destination_endpoint)
        sock.setblocking(False)
        return sock

    sos = UDPOutputSession(
        specifier=OutputSessionSpecifier(MessageDataSpecifier(3210), None),
        payload_metadata=PayloadMetadata(1024),
        mtu=11,
        multiplier=1,
        sock=make_sock(),
        loop=asyncio.get_event_loop(),
        finalizer=do_finalize,
    )

    assert sos.specifier == OutputSessionSpecifier(MessageDataSpecifier(3210), None)
    assert sos.destination_node_id is None
    assert sos.payload_metadata == PayloadMetadata(1024)
    assert sos.sample_statistics() == SessionStatistics()

    assert run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.NOMINAL,
                transfer_id=12340,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            loop.time() + 10.0,
        )
    )

    rx_data, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    assert rx_data == (
        b"\x00\x04\x00\x00\x00\x00\x00\x8040\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"one"
        + b"two"
        + b"three"
    )
    with raises(socket_.timeout):
        sock_rx.recvfrom(1000)

    last_feedback: typing.Optional[Feedback] = None

    def feedback_handler(feedback: Feedback) -> None:
        nonlocal last_feedback
        last_feedback = feedback

    sos.enable_feedback(feedback_handler)

    assert last_feedback is None
    assert run_until_complete(
        sos.send(
            Transfer(timestamp=ts, priority=Priority.NOMINAL, transfer_id=12340, fragmented_payload=[]),
            loop.time() + 10.0,
        )
    )
    assert last_feedback is not None
    assert last_feedback.original_transfer_timestamp == ts
    assert check_timestamp(last_feedback.first_frame_transmission_timestamp)

    sos.disable_feedback()
    sos.disable_feedback()  # Idempotency check

    _, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    with raises(socket_.timeout):
        sock_rx.recvfrom(1000)

    assert sos.sample_statistics() == SessionStatistics(transfers=2, frames=2, payload_bytes=11, errors=0, drops=0)

    assert sos.socket.fileno() >= 0
    assert not finalized
    sos.close()
    assert finalized
    assert sos.socket.fileno() < 0  # The socket is supposed to be disposed of.
    finalized = False

    # Multi-frame with multiplication
    sos = UDPOutputSession(
        specifier=OutputSessionSpecifier(ServiceDataSpecifier(321, ServiceDataSpecifier.Role.REQUEST), 2222),
        payload_metadata=PayloadMetadata(1024),
        mtu=10,
        multiplier=2,
        sock=make_sock(),
        loop=asyncio.get_event_loop(),
        finalizer=do_finalize,
    )
    assert run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.OPTIONAL,
                transfer_id=54321,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            loop.time() + 10.0,
        )
    )
    data_main_a, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    data_main_b, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    data_redundant_a, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    data_redundant_b, endpoint = sock_rx.recvfrom(1000)
    assert endpoint[0] == "127.100.0.2"
    with raises(socket_.timeout):
        sock_rx.recvfrom(1000)

    print("data_main_a", data_main_a)
    print("data_main_b", data_main_b)
    print("data_redundant_a", data_redundant_a)
    print("data_redundant_b", data_redundant_b)

    assert data_main_a == data_redundant_a
    assert data_main_b == data_redundant_b
    assert data_main_a == (
        b"\x00\x07\x00\x00\x00\x00\x00\x001\xd4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"one"
        + b"two"
        + b"three"[:-1]
    )
    assert data_main_b == (
        b"\x00\x07\x00\x00\x01\x00\x00\x801\xd4\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        + b"e"
        + pyuavcan.transport.commons.crc.CRC32C.new(b"one", b"two", b"three").value_as_bytes
    )

    sos.socket.close()  # This is to prevent resource warning
    sos = UDPOutputSession(
        specifier=OutputSessionSpecifier(ServiceDataSpecifier(321, ServiceDataSpecifier.Role.REQUEST), 2222),
        payload_metadata=PayloadMetadata(1024),
        mtu=10,
        multiplier=1,
        sock=make_sock(),
        loop=asyncio.get_event_loop(),
        finalizer=do_finalize,
    )

    # Induced timeout
    assert not run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.NOMINAL,
                transfer_id=12340,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            loop.time() - 0.1,  # Expired on arrival
        )
    )

    assert sos.sample_statistics() == SessionStatistics(
        transfers=0, frames=0, payload_bytes=0, errors=0, drops=2  # Because multiframe
    )

    # Induced failure
    sos.socket.close()
    with raises(OSError):
        assert not run_until_complete(
            sos.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.NOMINAL,
                    transfer_id=12340,
                    fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
                ),
                loop.time() + 10.0,
            )
        )

    assert sos.sample_statistics() == SessionStatistics(transfers=0, frames=0, payload_bytes=0, errors=1, drops=2)

    assert not finalized
    sos.close()
    assert finalized
    sos.close()  # Idempotency

    with raises(pyuavcan.transport.ResourceClosedError):
        run_until_complete(
            sos.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.NOMINAL,
                    transfer_id=12340,
                    fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
                ),
                loop.time() + 10.0,
            )
        )

    sock_rx.close()


def _unittest_output_session_no_listener() -> None:
    """
    Test the Windows-specific corner case. Should be handled identically on all platforms.
    """
    from pyuavcan.transport import OutputSessionSpecifier, MessageDataSpecifier, Priority
    from pyuavcan.transport import PayloadMetadata, Feedback, Transfer

    ts = Timestamp.now()
    loop = asyncio.get_event_loop()
    run_until_complete = loop.run_until_complete

    def make_sock() -> socket_.socket:
        sock = socket_.socket(socket_.AF_INET, socket_.SOCK_DGRAM)
        sock.bind(("127.100.0.2", 0))
        sock.connect(("239.0.1.2", 33333))  # There is no listener on this endpoint.
        sock.setsockopt(socket_.IPPROTO_IP, socket_.IP_MULTICAST_IF, socket_.inet_aton("127.100.0.2"))
        sock.setblocking(False)
        return sock

    sos = UDPOutputSession(
        specifier=OutputSessionSpecifier(MessageDataSpecifier(3210), None),
        payload_metadata=PayloadMetadata(1024),
        mtu=11,
        multiplier=1,
        sock=make_sock(),
        loop=asyncio.get_event_loop(),
        finalizer=lambda: None,
    )
    assert run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.NOMINAL,
                transfer_id=12340,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            loop.time() + 10.0,
        )
    )
    sos.close()

    # Multi-frame with multiplication and feedback
    last_feedback: typing.Optional[Feedback] = None

    def feedback_handler(feedback: Feedback) -> None:
        nonlocal last_feedback
        assert last_feedback is None, "Unexpected feedback"
        last_feedback = feedback

    sos = UDPOutputSession(
        specifier=OutputSessionSpecifier(ServiceDataSpecifier(321, ServiceDataSpecifier.Role.REQUEST), 2222),
        payload_metadata=PayloadMetadata(1024),
        mtu=10,
        multiplier=2,
        sock=make_sock(),
        loop=asyncio.get_event_loop(),
        finalizer=lambda: None,
    )
    sos.enable_feedback(feedback_handler)
    assert last_feedback is None
    assert run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.OPTIONAL,
                transfer_id=54321,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            loop.time() + 10.0,
        )
    )
    print("last_feedback:", last_feedback)
    assert isinstance(last_feedback, UDPFeedback)
    # Ensure that the timestamp is populated even if the error suppression logic is activated.
    assert last_feedback.original_transfer_timestamp == ts
    assert Timestamp.now().monotonic >= last_feedback.first_frame_transmission_timestamp.monotonic >= ts.monotonic
    assert Timestamp.now().system >= last_feedback.first_frame_transmission_timestamp.system >= ts.system

    sos.close()



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import sys
import errno
import typing
import socket
import logging
from ipaddress import IPv4Address, IPV4LENGTH, ip_network
import pyuavcan
from pyuavcan.transport import MessageDataSpecifier, ServiceDataSpecifier, UnsupportedSessionConfigurationError
from pyuavcan.transport import InvalidMediaConfigurationError
from ._socket_factory import SocketFactory, Sniffer
from ._endpoint_mapping import SUBJECT_PORT, IP_ADDRESS_NODE_ID_MASK, service_data_specifier_to_udp_port
from ._endpoint_mapping import node_id_to_unicast_ip, message_data_specifier_to_multicast_group
from ._link_layer import LinkLayerCapture, LinkLayerSniffer


_logger = logging.getLogger(__name__)


class IPv4SocketFactory(SocketFactory):
    """
    In IPv4 networks, the node-ID of zero may not be usable because it represents the subnet address;
    a node-ID that maps to the broadcast address for the subnet is unavailable.
    """

    def __init__(self, local_ip_address: IPv4Address):
        if not isinstance(local_ip_address, IPv4Address):  # pragma: no cover
            raise TypeError(f"Unexpected IP address type: {type(local_ip_address)}")
        self._local = local_ip_address

    @property
    def max_nodes(self) -> int:
        return IP_ADDRESS_NODE_ID_MASK  # The maximum may not be available because it may be the broadcast address.

    @property
    def local_ip_address(self) -> IPv4Address:
        return self._local

    def make_output_socket(
        self, remote_node_id: typing.Optional[int], data_specifier: pyuavcan.transport.DataSpecifier
    ) -> socket.socket:
        _logger.debug(
            "%r: Constructing new output socket for remote node %s and %s", self, remote_node_id, data_specifier
        )
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        s.setblocking(False)
        try:
            # Output sockets shall be bound, too, in order to ensure that outgoing packets have the correct
            # source IP address specified. This is particularly important for localhost; an unbound socket
            # there emits all packets from 127.0.0.1 which is certainly not what we need.
            s.bind((str(self._local), 0))  # Bind to an ephemeral port.
        except OSError as ex:
            s.close()
            if ex.errno == errno.EADDRNOTAVAIL:
                raise InvalidMediaConfigurationError(
                    f"Bad IP configuration: cannot bind output socket to {self._local} [{errno.errorcode[ex.errno]}]"
                ) from None
            raise  # pragma: no cover

        if isinstance(data_specifier, MessageDataSpecifier):
            if remote_node_id is not None:
                s.close()
                raise UnsupportedSessionConfigurationError("Unicast message transfers are not defined.")
            # Merely binding is not enough for multicast sockets. We also have to configure IP_MULTICAST_IF.
            # https://tldp.org/HOWTO/Multicast-HOWTO-6.html
            # https://stackoverflow.com/a/26988214/1007777
            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, self._local.packed)
            s.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, IPv4SocketFactory.MULTICAST_TTL)
            remote_ip = message_data_specifier_to_multicast_group(self._local, data_specifier)
            remote_port = SUBJECT_PORT
        elif isinstance(data_specifier, ServiceDataSpecifier):
            if remote_node_id is None:
                s.close()
                raise UnsupportedSessionConfigurationError("Broadcast service transfers are not defined.")
            remote_ip = node_id_to_unicast_ip(self._local, remote_node_id)
            remote_port = service_data_specifier_to_udp_port(data_specifier)
        else:
            assert False

        s.connect((str(remote_ip), remote_port))
        _logger.debug("%r: New output %r connected to remote node %r", self, s, remote_node_id)
        return s

    def make_input_socket(self, data_specifier: pyuavcan.transport.DataSpecifier) -> socket.socket:
        _logger.debug("%r: Constructing new input socket for %s", self, data_specifier)
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        s.setblocking(False)
        # Allow other applications to use the same UAVCAN port as well.
        # These options shall be set before the socket is bound.
        # https://stackoverflow.com/questions/14388706/how-do-so-reuseaddr-and-so-reuseport-differ/14388707#14388707
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if sys.platform.startswith("linux"):  # pragma: no branch
            # This is expected to be useful for unicast inputs only.
            # https://stackoverflow.com/a/14388707/1007777
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

        if isinstance(data_specifier, MessageDataSpecifier):
            multicast_ip = message_data_specifier_to_multicast_group(self._local, data_specifier)
            multicast_port = SUBJECT_PORT
            if sys.platform.startswith("linux"):
                # Binding to the multicast group address is necessary on GNU/Linux: https://habr.com/ru/post/141021/
                s.bind((str(multicast_ip), multicast_port))
            else:
                # Binding to a multicast address is not allowed on Windows, and it is not necessary there. Error is:
                #   OSError: [WinError 10049] The requested address is not valid in its context
                s.bind(("", multicast_port))
            try:
                # Note that using INADDR_ANY in IP_ADD_MEMBERSHIP doesn't actually mean "any",
                # it means "choose one automatically"; see https://tldp.org/HOWTO/Multicast-HOWTO-6.html
                # This is why we have to specify the interface explicitly here.
                s.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, multicast_ip.packed + self._local.packed)
            except OSError as ex:
                s.close()
                if ex.errno in (errno.EADDRNOTAVAIL, errno.ENODEV):
                    raise InvalidMediaConfigurationError(
                        f"Could not register multicast group membership {multicast_ip} via {self._local} using {s} "
                        f"[{errno.errorcode[ex.errno]}]"
                    ) from None
                raise  # pragma: no cover
        elif isinstance(data_specifier, ServiceDataSpecifier):
            local_port = service_data_specifier_to_udp_port(data_specifier)
            try:
                s.bind((str(self._local), local_port))
            except OSError as ex:
                s.close()
                if ex.errno in (errno.EADDRNOTAVAIL, errno.ENODEV):
                    raise InvalidMediaConfigurationError(
                        f"Could not bind input service socket to {self._local}:{local_port} "
                        f"[{errno.errorcode[ex.errno]}]"
                    ) from None
                raise  # pragma: no cover
        else:
            assert False
        _logger.debug("%r: New input %r", self, s)
        return s

    def make_sniffer(self, handler: typing.Callable[[LinkLayerCapture], None]) -> SnifferIPv4:
        return SnifferIPv4(self._local, handler)


class SnifferIPv4(Sniffer):
    def __init__(self, local_ip_address: IPv4Address, handler: typing.Callable[[LinkLayerCapture], None]) -> None:
        netmask_width = IPV4LENGTH - IP_ADDRESS_NODE_ID_MASK.bit_length()
        subnet = ip_network(f"{local_ip_address}/{netmask_width}", strict=False)
        filter_expression = f"udp and src net {subnet}"
        _logger.debug("Constructed BPF filter expression: %r", filter_expression)
        self._link_layer = LinkLayerSniffer(filter_expression, handler)

    def close(self) -> None:
        self._link_layer.close()

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self._link_layer)



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import dataclasses
import pyuavcan
import pyuavcan.transport.can
from pyuavcan.transport import Trace, TransferTrace, AlienSessionSpecifier, AlienTransferMetadata, Capture
from pyuavcan.transport import AlienTransfer, TransferFrom, Timestamp, Priority
from ._session import TransferReassemblyErrorID, TransferReassembler
from .media import DataFrame
from ._frame import UAVCANFrame
from ._identifier import CANID


@dataclasses.dataclass(frozen=True)
class CANCapture(Capture):
    """
    See :meth:`pyuavcan.transport.can.CANTransport.begin_capture` for details.
    """

    frame: DataFrame

    own: bool
    """
    True if the captured frame was sent by the local transport instance.
    False if it was received from the bus.
    """

    def parse(self) -> typing.Optional[typing.Tuple[AlienSessionSpecifier, Priority, UAVCANFrame]]:
        uf = UAVCANFrame.parse(self.frame)
        if not uf:
            return None
        ci = CANID.parse(self.frame.identifier)
        if not ci:
            return None
        ss = AlienSessionSpecifier(
            source_node_id=ci.source_node_id,
            destination_node_id=ci.get_destination_node_id(),
            data_specifier=ci.data_specifier,
        )
        return ss, ci.priority, uf

    def __repr__(self) -> str:
        direction = "tx" if self.own else "rx"
        return pyuavcan.util.repr_attributes(self, self.timestamp, direction, self.frame)

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.can.CANTransport]:
        return pyuavcan.transport.can.CANTransport


@dataclasses.dataclass(frozen=True)
class CANErrorTrace(pyuavcan.transport.ErrorTrace):
    error: TransferReassemblyErrorID


class CANTracer(pyuavcan.transport.Tracer):
    """
    The CAN tracer does not differentiate between RX and TX frames, they are treated uniformly.
    Return types from :meth:`update`:

    - :class:`pyuavcan.transport.TransferTrace`
    - :class:`CANErrorTrace`
    """

    def __init__(self) -> None:
        self._sessions: typing.Dict[AlienSessionSpecifier, _AlienSession] = {}

    def update(self, cap: Capture) -> typing.Optional[Trace]:
        if not isinstance(cap, CANCapture):
            return None
        parsed = cap.parse()
        if not parsed:
            return None
        ss, prio, frame = parsed
        if ss.source_node_id is not None:
            return self._get_session(ss).update(cap.timestamp, prio, frame)
        # Anonymous transfer -- no reconstruction needed, no session.
        return TransferTrace(
            cap.timestamp,
            AlienTransfer(AlienTransferMetadata(prio, frame.transfer_id, ss), [frame.padded_payload]),
            0.0,
        )

    def _get_session(self, specifier: AlienSessionSpecifier) -> _AlienSession:
        try:
            return self._sessions[specifier]
        except KeyError:
            self._sessions[specifier] = _AlienSession(specifier)
        return self._sessions[specifier]


class _AlienSession:
    _MAX_INTERVAL = 1.0
    _TID_TIMEOUT_MULTIPLIER = 2.0  # TID = 2*interval as suggested in the Specification.
    _EXTENT_BYTES = 2 ** 32

    def __init__(self, specifier: AlienSessionSpecifier) -> None:
        assert specifier.source_node_id is not None
        self._specifier = specifier
        self._reassembler = TransferReassembler(
            source_node_id=specifier.source_node_id, extent_bytes=_AlienSession._EXTENT_BYTES
        )
        self._last_transfer_monotonic: float = 0.0
        self._interval = float(_AlienSession._MAX_INTERVAL)

    def update(self, timestamp: Timestamp, priority: Priority, frame: UAVCANFrame) -> typing.Optional[Trace]:
        tid_timeout = self.transfer_id_timeout
        tr = self._reassembler.process_frame(timestamp, priority, frame, int(tid_timeout * 1e9))
        if tr is None:
            return None
        if isinstance(tr, TransferReassemblyErrorID):
            return CANErrorTrace(timestamp=timestamp, error=tr)

        assert isinstance(tr, TransferFrom)
        meta = AlienTransferMetadata(tr.priority, tr.transfer_id, self._specifier)
        out = TransferTrace(timestamp, AlienTransfer(meta, tr.fragmented_payload), tid_timeout)

        # Update the transfer interval for automatic transfer-ID timeout deduction.
        delta = float(tr.timestamp.monotonic) - self._last_transfer_monotonic
        delta = min(_AlienSession._MAX_INTERVAL, max(0.0, delta))
        self._interval = (self._interval + delta) * 0.5
        self._last_transfer_monotonic = float(tr.timestamp.monotonic)

        return out

    @property
    def transfer_id_timeout(self) -> float:
        """
        The current value of the auto-deduced transfer-ID timeout.
        It is automatically adjusted whenever a new transfer is received.
        """
        return self._interval * _AlienSession._TID_TIMEOUT_MULTIPLIER


# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------


def _unittest_can_capture() -> None:
    from pyuavcan.transport import MessageDataSpecifier
    from .media import FrameFormat
    from ._identifier import MessageCANID

    ts = Timestamp.now()
    payload = bytearray(b"123\x0A")
    cap = CANCapture(
        ts,
        DataFrame(
            FrameFormat.EXTENDED,
            MessageCANID(Priority.SLOW, 42, 3210).compile([memoryview(payload)]),
            payload,
        ),
        own=True,
    )
    print(cap)
    parsed = cap.parse()
    assert parsed is not None
    ss, prio, uf = parsed
    assert ss.source_node_id == 42
    assert ss.destination_node_id is None
    assert isinstance(ss.data_specifier, MessageDataSpecifier)
    assert ss.data_specifier.subject_id == 3210
    assert prio == Priority.SLOW
    assert uf.transfer_id == 0x0A
    assert uf.padded_payload == b"123"
    assert not uf.start_of_transfer
    assert not uf.end_of_transfer
    assert not uf.toggle_bit

    # Invalid CAN ID
    assert None is CANCapture(ts, DataFrame(FrameFormat.BASE, 123, payload), own=True).parse()

    # Invalid CAN payload
    assert (
        None
        is CANCapture(
            ts,
            DataFrame(FrameFormat.EXTENDED, MessageCANID(Priority.SLOW, 42, 3210).compile([]), bytearray()),
            own=True,
        ).parse()
    )


def _unittest_can_alien_session() -> None:
    from pytest import approx
    from pyuavcan.transport import MessageDataSpecifier
    from ._identifier import MessageCANID

    ts = Timestamp.now()
    can_identifier = MessageCANID(Priority.SLOW, 42, 3210).compile([])

    def frm(
        padded_payload: typing.Union[bytes, str],
        transfer_id: int,
        start_of_transfer: bool,
        end_of_transfer: bool,
        toggle_bit: bool,
    ) -> UAVCANFrame:
        return UAVCANFrame(
            identifier=can_identifier,
            padded_payload=memoryview(padded_payload if isinstance(padded_payload, bytes) else padded_payload.encode()),
            transfer_id=transfer_id,
            start_of_transfer=start_of_transfer,
            end_of_transfer=end_of_transfer,
            toggle_bit=toggle_bit,
        )

    spec = AlienSessionSpecifier(42, None, MessageDataSpecifier(3210))
    ses = _AlienSession(spec)

    # Valid multi-frame (test data copy-posted from the reassembler test).
    assert None is ses.update(ts, Priority.HIGH, frm(b"\x00\x01\x02\x03\x04\x05\x06", 11, True, False, True))
    assert None is ses.update(ts, Priority.HIGH, frm(b"\x07\x08\x09\x0a\x0b\x0c\x0d", 11, False, False, False))
    assert None is ses.update(ts, Priority.HIGH, frm(b"\x0e\x0f\x10\x11\x12\x13\x14", 11, False, False, True))
    assert None is ses.update(ts, Priority.HIGH, frm(b"\x15\x16\x17\x18\x19\x1a\x1b", 11, False, False, False))
    tr = ses.update(ts, Priority.HIGH, frm(b"\x1c\x1d\x35\x54", 11, False, True, True))
    assert isinstance(tr, TransferTrace)
    assert list(tr.transfer.fragmented_payload) == [
        b"\x00\x01\x02\x03\x04\x05\x06",
        b"\x07\x08\x09\x0a\x0b\x0c\x0d",
        b"\x0e\x0f\x10\x11\x12\x13\x14",
        b"\x15\x16\x17\x18\x19\x1a\x1b",
        b"\x1c\x1d",  # CRC stripped
    ]
    assert tr.transfer.metadata.priority == Priority.HIGH
    assert tr.transfer.metadata.transfer_id == 11
    assert tr.transfer.metadata.session_specifier.source_node_id == 42
    assert tr.transfer.metadata.session_specifier.destination_node_id is None
    assert isinstance(tr.transfer.metadata.session_specifier.data_specifier, MessageDataSpecifier)
    assert tr.transfer.metadata.session_specifier.data_specifier.subject_id == 3210
    assert tr.timestamp == ts
    assert tr.transfer_id_timeout == approx(2.0)  # Default value.

    # Missed start of transfer.
    tr = ses.update(ts, Priority.HIGH, frm(b"123456", 2, False, False, False))
    assert isinstance(tr, CANErrorTrace)

    # Valid single-frame; TID timeout updated.
    tr = ses.update(ts, Priority.LOW, frm(b"\x00\x01\x02\x03\x04\x05\x06", 12, True, True, True))
    assert isinstance(tr, TransferTrace)
    assert tr.transfer.metadata.priority == Priority.LOW
    assert tr.transfer.metadata.transfer_id == 12
    assert tr.transfer.metadata.session_specifier.source_node_id == 42
    assert tr.transfer.metadata.session_specifier.destination_node_id is None
    assert isinstance(tr.transfer.metadata.session_specifier.data_specifier, MessageDataSpecifier)
    assert tr.transfer.metadata.session_specifier.data_specifier.subject_id == 3210
    assert tr.timestamp == ts
    assert ses.transfer_id_timeout == approx(1.0)  # Shrunk twice because we're using the same timestamp here.


def _unittest_can_tracer() -> None:
    from .media import FrameFormat
    from ._identifier import MessageCANID

    ts = Timestamp.now()
    tracer = CANTracer()

    # Foreign capture ignored.
    assert None is tracer.update(Capture(ts))

    # Valid transfers.
    cap = CANCapture(
        ts,
        DataFrame(
            FrameFormat.EXTENDED,
            MessageCANID(Priority.FAST, 42, 3210).compile([]),
            bytearray(b"123\xFF"),
        ),
        own=True,
    )
    tr = tracer.update(cap)
    assert isinstance(tr, TransferTrace)
    assert tr.timestamp == ts
    assert tr.transfer.metadata.transfer_id == 31
    assert tr.transfer.metadata.priority == Priority.FAST
    assert tr.transfer.metadata.session_specifier.source_node_id == 42

    cap = CANCapture(
        ts,
        DataFrame(
            FrameFormat.EXTENDED,
            MessageCANID(Priority.SLOW, 42, 3210).compile([]),
            bytearray(b"123\xE0"),
        ),
        own=False,  # Direction is ignored.
    )
    tr = tracer.update(cap)
    assert isinstance(tr, TransferTrace)
    assert tr.timestamp == ts
    assert tr.transfer.metadata.transfer_id == 0
    assert tr.transfer.metadata.priority == Priority.SLOW
    assert tr.transfer.metadata.session_specifier.source_node_id == 42

    cap = CANCapture(
        ts,
        DataFrame(
            FrameFormat.EXTENDED,
            MessageCANID(Priority.SLOW, None, 3210).compile([]),
            bytearray(b"123\xE0"),
        ),
        own=False,  # Direction is ignored.
    )
    tr = tracer.update(cap)
    assert isinstance(tr, TransferTrace)
    assert tr.timestamp == ts
    assert tr.transfer.metadata.transfer_id == 0
    assert tr.transfer.metadata.priority == Priority.SLOW
    assert tr.transfer.metadata.session_specifier.source_node_id is None

    # Invalid captured frame.
    assert None is tracer.update(CANCapture(ts, DataFrame(FrameFormat.BASE, 123, bytearray(b"")), own=False))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import enum
import time
import errno
import typing
import socket
import struct
import select
import asyncio
import logging
import threading
import contextlib
import pyuavcan.transport
from pyuavcan.transport import Timestamp
from pyuavcan.transport.can.media import Media, Envelope, FilterConfiguration, FrameFormat
from pyuavcan.transport.can.media import DataFrame

# Disable unused ignore warning for this file only because there appears to be no other way to make MyPy
# accept this file both on Windows and GNU/Linux.
# mypy: warn_unused_ignores=False


_logger = logging.getLogger(__name__)


class SocketCANMedia(Media):
    """
    This media implementation provides a simple interface for the standard Linux SocketCAN media layer.
    If you are testing with a virtual CAN bus and you need CAN FD, you may need to enable it manually
    (https://stackoverflow.com/questions/36568167/can-fd-support-for-virtual-can-vcan-on-socketcan);
    otherwise, you may observe errno 90 "Message too long". Configuration example::

        ip link set vcan0 mtu 72

    SocketCAN documentation: https://www.kernel.org/doc/Documentation/networking/can.txt
    """

    def __init__(self, iface_name: str, mtu: int, loop: typing.Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        CAN Classic/FD is selected automatically based on the MTU. It is not possible to use CAN FD with MTU of 8 bytes.

        :param iface_name: E.g., ``can0``.

        :param mtu: The maximum data field size in bytes. CAN FD is used if this value > 8, Classic CAN otherwise.
            This value must belong to Media.VALID_MTU_SET.

        :param loop: The event loop to use. Defaults to :func:`asyncio.get_event_loop`.
        """
        self._mtu = int(mtu)
        if self._mtu not in self.VALID_MTU_SET:
            raise ValueError(f"Invalid MTU: {self._mtu} not in {self.VALID_MTU_SET}")

        self._iface_name = str(iface_name)
        self._loop = loop if loop is not None else asyncio.get_event_loop()

        self._is_fd = self._mtu > _NativeFrameDataCapacity.CAN_CLASSIC
        self._native_frame_data_capacity = int(
            {
                False: _NativeFrameDataCapacity.CAN_CLASSIC,
                True: _NativeFrameDataCapacity.CAN_FD,
            }[self._is_fd]
        )
        self._native_frame_size = _FRAME_HEADER_STRUCT.size + self._native_frame_data_capacity

        self._sock = _make_socket(iface_name, can_fd=self._is_fd)
        self._ctl_main, self._ctl_worker = socket.socketpair()  # This is used for controlling the worker thread.
        self._closed = False
        self._maybe_thread: typing.Optional[threading.Thread] = None
        self._loopback_enabled = False

        self._ancillary_data_buffer_size = socket.CMSG_SPACE(_TIMEVAL_STRUCT.size)  # Used for recvmsg()

        super().__init__()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    @property
    def interface_name(self) -> str:
        return self._iface_name

    @property
    def mtu(self) -> int:
        return self._mtu

    @property
    def number_of_acceptance_filters(self) -> int:
        """
        512 for SocketCAN.

        - https://github.com/torvalds/linux/blob/9c7db5004280767566e91a33445bf93aa479ef02/net/can/af_can.c#L327-L348
        - https://github.com/torvalds/linux/blob/54dee406374ce8adb352c48e175176247cb8db7c/include/uapi/linux/can.h#L200
        """
        return 512

    def start(self, handler: Media.ReceivedFramesHandler, no_automatic_retransmission: bool) -> None:
        if self._maybe_thread is None:
            self._maybe_thread = threading.Thread(
                target=self._thread_function, name=str(self), args=(handler,), daemon=True
            )
            self._maybe_thread.start()
            if no_automatic_retransmission:
                _logger.info("%s non-automatic retransmission is not supported", self)
        else:
            raise RuntimeError("The RX frame handler is already set up")

    def configure_acceptance_filters(self, configuration: typing.Sequence[FilterConfiguration]) -> None:
        if self._closed:
            raise pyuavcan.transport.ResourceClosedError(repr(self))
        _logger.info(
            "%s FIXME: acceptance filter configuration is not yet implemented; please submit patches! "
            "Requested configuration: %s",
            self,
            ", ".join(map(str, configuration)),
        )

    async def send(self, frames: typing.Iterable[Envelope], monotonic_deadline: float) -> int:
        num_sent = 0
        for f in frames:
            if self._closed:
                raise pyuavcan.transport.ResourceClosedError(repr(self))
            self._set_loopback_enabled(f.loopback)
            try:
                await asyncio.wait_for(
                    self._loop.sock_sendall(self._sock, self._compile_native_frame(f.frame)),
                    timeout=monotonic_deadline - self._loop.time(),
                )
            except asyncio.TimeoutError:
                break
            else:
                num_sent += 1
        return num_sent

    def close(self) -> None:
        try:
            self._closed = True
            if self._ctl_main.fileno() >= 0:  # Ignore if already closed.
                self._ctl_main.send(b"stop")  # The actual data is irrelevant, we just need it to unblock the select().
            if self._maybe_thread:
                self._maybe_thread.join(timeout=_SELECT_TIMEOUT)
                self._maybe_thread = None
        finally:
            self._sock.close()  # These are expected to be idempotent.
            self._ctl_worker.close()
            self._ctl_main.close()

    def _thread_function(self, handler: Media.ReceivedFramesHandler) -> None:
        def handler_wrapper(frs: typing.Sequence[typing.Tuple[Timestamp, Envelope]]) -> None:
            try:
                if not self._closed:  # Don't call after closure to prevent race conditions and use-after-close.
                    handler(frs)
            except Exception as exc:
                _logger.exception("%s: Unhandled exception in the receive handler: %s; lost frames: %s", self, exc, frs)

        while not self._closed:
            try:
                (
                    read_ready,
                    _,
                    _,
                ) = select.select((self._sock, self._ctl_worker), (), (), _SELECT_TIMEOUT)
                ts_mono_ns = time.monotonic_ns()

                if self._sock in read_ready:
                    frames: typing.List[typing.Tuple[Timestamp, Envelope]] = []
                    try:
                        while True:
                            frames.append(self._read_frame(ts_mono_ns))
                    except OSError as ex:
                        if ex.errno != errno.EAGAIN:
                            raise
                    self._loop.call_soon_threadsafe(handler_wrapper, frames)

                if self._ctl_worker in read_ready:
                    if self._ctl_worker.recv(1):  # pragma: no branch
                        break
            except Exception as ex:  # pragma: no cover
                if self._sock.fileno() < 0 or self._ctl_worker.fileno() < 0 or self._ctl_main.fileno() < 0:
                    self._closed = True
                _logger.exception("%s thread failure: %s", self, ex)
                time.sleep(1)  # Is this an adequate failure management strategy?

        self._closed = True
        _logger.debug("%s thread is about to exit", self)

    def _read_frame(self, ts_mono_ns: int) -> typing.Tuple[Timestamp, Envelope]:
        while True:
            data, ancdata, msg_flags, _addr = self._sock.recvmsg(
                self._native_frame_size, self._ancillary_data_buffer_size
            )
            assert msg_flags & socket.MSG_TRUNC == 0, "The data buffer is not large enough"
            assert msg_flags & socket.MSG_CTRUNC == 0, "The ancillary data buffer is not large enough"

            loopback = bool(msg_flags & socket.MSG_CONFIRM)
            ts_system_ns = 0
            for cmsg_level, cmsg_type, cmsg_data in ancdata:
                if cmsg_level == socket.SOL_SOCKET and cmsg_type == _SO_TIMESTAMP:
                    sec, usec = _TIMEVAL_STRUCT.unpack(cmsg_data)
                    ts_system_ns = (sec * 1_000_000 + usec) * 1000
                else:
                    assert False, f"Unexpected ancillary data: {cmsg_level}, {cmsg_type}, {cmsg_data!r}"

            assert ts_system_ns > 0, "Missing the timestamp; does the driver support timestamping?"
            timestamp = Timestamp(system_ns=ts_system_ns, monotonic_ns=ts_mono_ns)
            out = SocketCANMedia._parse_native_frame(data)
            if out is not None:
                return timestamp, Envelope(out, loopback=loopback)

    def _compile_native_frame(self, source: DataFrame) -> bytes:
        flags = _CANFD_BRS if self._is_fd else 0
        ident = source.identifier | (_CAN_EFF_FLAG if source.format == FrameFormat.EXTENDED else 0)
        header = _FRAME_HEADER_STRUCT.pack(ident, len(source.data), flags)
        out = header + source.data.ljust(self._native_frame_data_capacity, b"\x00")
        assert len(out) == self._native_frame_size
        return out

    @staticmethod
    def _parse_native_frame(source: bytes) -> typing.Optional[DataFrame]:
        header_size = _FRAME_HEADER_STRUCT.size
        ident_raw, data_length, _flags = _FRAME_HEADER_STRUCT.unpack(source[:header_size])
        if (ident_raw & _CAN_RTR_FLAG) or (ident_raw & _CAN_ERR_FLAG):  # Unsupported format, ignore silently
            _logger.debug("Unsupported CAN frame dropped; raw SocketCAN ID is %08x", ident_raw)
            return None
        frame_format = FrameFormat.EXTENDED if ident_raw & _CAN_EFF_FLAG else FrameFormat.BASE
        data = source[header_size : header_size + data_length]
        assert len(data) == data_length
        ident = ident_raw & _CAN_EFF_MASK
        return DataFrame(frame_format, ident, bytearray(data))

    def _set_loopback_enabled(self, enable: bool) -> None:
        if enable != self._loopback_enabled:
            self._sock.setsockopt(socket.SOL_CAN_RAW, socket.CAN_RAW_RECV_OWN_MSGS, int(enable))  # type: ignore
            self._loopback_enabled = enable

    @staticmethod
    def list_available_interface_names() -> typing.Iterable[str]:
        import re
        import subprocess

        try:
            proc = subprocess.run("ip link show", check=True, timeout=1, text=True, shell=True, capture_output=True)
            return re.findall(r"\d+?: ([a-z0-9]+?): <[^>]*UP[^>]*>.*\n *link/can", proc.stdout)
        except Exception as ex:
            _logger.debug(
                "Could not scrape the output of `ip link show`, using the fallback method: %s", ex, exc_info=True
            )
            with open("/proc/net/dev") as f:
                out = [line.split(":")[0].strip() for line in f if ":" in line and "can" in line]
            return sorted(out, key=lambda x: "can" in x, reverse=True)


class _NativeFrameDataCapacity(enum.IntEnum):
    CAN_CLASSIC = 8
    CAN_FD = 64


_SELECT_TIMEOUT = 1.0


# struct can_frame {
#     canid_t can_id;  /* 32 bit CAN_ID + EFF/RTR/ERR flags */
#     __u8    can_dlc; /* data length code: 0 .. 8 */
#     __u8    data[8] __attribute__((aligned(8)));
# };
# struct canfd_frame {
#     canid_t can_id;  /* 32 bit CAN_ID + EFF/RTR/ERR flags */
#     __u8    len;     /* frame payload length in byte */
#     __u8    flags;   /* additional flags for CAN FD */
#     __u8    __res0;  /* reserved / padding */
#     __u8    __res1;  /* reserved / padding */
#     __u8    data[CANFD_MAX_DLEN] __attribute__((aligned(8)));
# };
_FRAME_HEADER_STRUCT = struct.Struct("=IBB2x")  # Using standard size because the native definition relies on stdint.h
_TIMEVAL_STRUCT = struct.Struct("@Ll")  # Using native size because the native definition uses plain integers

# From the Linux kernel; not exposed via the Python's socket module
_SO_TIMESTAMP = 29

_CANFD_BRS = 1

_CAN_EFF_FLAG = 0x80000000
_CAN_RTR_FLAG = 0x40000000
_CAN_ERR_FLAG = 0x20000000

_CAN_EFF_MASK = 0x1FFFFFFF


def _make_socket(iface_name: str, can_fd: bool) -> socket.SocketType:
    s = socket.socket(socket.PF_CAN, socket.SOCK_RAW, socket.CAN_RAW)  # type: ignore
    try:
        s.bind((iface_name,))
        s.setsockopt(socket.SOL_SOCKET, _SO_TIMESTAMP, 1)  # timestamping
        if can_fd:
            s.setsockopt(socket.SOL_CAN_RAW, socket.CAN_RAW_FD_FRAMES, 1)  # type: ignore

        s.setblocking(False)

        if 0 != s.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR):
            raise OSError("Could not configure the socket: getsockopt(SOL_SOCKET, SO_ERROR) != 0")
    except BaseException:
        with contextlib.suppress(Exception):
            s.close()
        raise

    return s



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import warnings
import dataclasses
import pyuavcan.util
from ._transfer import Transfer, TransferFrom
from ._timestamp import Timestamp
from ._data_specifier import DataSpecifier
from ._payload_metadata import PayloadMetadata


class Feedback(abc.ABC):
    """
    Abstract output transfer feedback for transmission timestamping.
    If feedback is enabled for an output session, an instance of this class is delivered back to the application
    via a callback soon after the first frame of the transfer is emitted.

    The upper layers can match a feedback object with its transfer by the transfer creation timestamp.
    """

    @property
    @abc.abstractmethod
    def original_transfer_timestamp(self) -> Timestamp:
        """
        This is the timestamp value of the original outgoing transfer object;
        normally it is the transfer creation timestamp.
        This value can be used by the upper layers to match each transmitted transfer with its transmission timestamp.
        Why do we use timestamp for matching? This is because:

        - The priority is rarely unique, hence unfit for matching.

        - Transfer-ID may be modified by the transport layer by computing its modulus, which is difficult to
          reliably account for in the application, especially in heterogeneous redundant transports.

        - The fragmented payload may contain references to the actual memory of the serialized object, meaning
          that it may actually change after the object is transmitted, also rendering it unfit for matching.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def first_frame_transmission_timestamp(self) -> Timestamp:
        """
        This is the best-effort estimate of the transmission timestamp.
        Transport implementations are not required to adhere to any specific accuracy goals.
        They may use either software or hardware timestamping under the hood,
        depending on the capabilities of the underlying media driver.
        The timestamp of a multi-frame transfer is the timestamp of its first frame.
        The overall TX latency can be computed by subtracting the original transfer timestamp from this value.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self,
            original_transfer_timestamp=self.original_transfer_timestamp,
            first_frame_transmission_timestamp=self.first_frame_transmission_timestamp,
        )


@dataclasses.dataclass(frozen=True)
class SessionSpecifier:
    """
    This dataclass models the session specifier (https://forum.uavcan.org/t/alternative-transport-protocols/324)
    except that we assume that one end of the session terminates at the local node.
    There are specializations for input and output sessions with additional logic,
    but they do not add extra data (because remember this class follows the protocol model definition).
    """

    data_specifier: DataSpecifier
    """
    See :class:`pyuavcan.transport.DataSpecifier`.
    """

    remote_node_id: typing.Optional[int]
    """
    If not None: output sessions are unicast to that node-ID, and input sessions ignore all transfers
    except those that originate from the specified remote node-ID.
    If None: output sessions are broadcast and input sessions are promiscuous.
    """

    def __post_init__(self) -> None:
        if self.remote_node_id is not None and self.remote_node_id < 0:
            raise ValueError(f"Invalid remote node-ID: {self.remote_node_id}")


@dataclasses.dataclass(frozen=True)
class InputSessionSpecifier(SessionSpecifier):
    """
    If the remote node-ID is set, this is a selective session (accept data from the specified remote node only);
    otherwise this is a promiscuous session (accept data from any node).
    """

    @property
    def is_promiscuous(self) -> bool:
        return self.remote_node_id is None


@dataclasses.dataclass(frozen=True)
class OutputSessionSpecifier(SessionSpecifier):
    """
    If the remote node-ID is set, this is a unicast session (use unicast transfers);
    otherwise this is a broadcast session (use broadcast transfers).
    The Specification v1.0 allows the following kinds of transfers:

    - Broadcast message transfers.
    - Unicast service transfers.

    Anything else is invalid per UAVCAN v1.0.
    A future version of the specification may add support for unicast messages for at least some transports.
    Here, we go ahead and assume that unicast message transfers are valid in general;
    it is up to a particular transport implementation to choose whether they are supported.
    Beware that this is a non-standard experimental protocol extension and it may be removed
    depending on how the next versions of the Specification evolve.
    You can influence that by leaving feedback at https://forum.uavcan.org.

    To summarize:

    +--------------------+--------------------------------------+---------------------------------------+
    |                    | Unicast                              | Broadcast                             |
    +====================+======================================+=======================================+
    | **Message**        | Experimental, may be allowed in v1.x | Allowed by Specification              |
    +--------------------+--------------------------------------+---------------------------------------+
    | **Service**        | Allowed by Specification             | Banned by Specification               |
    +--------------------+--------------------------------------+---------------------------------------+
    """

    def __post_init__(self) -> None:
        if isinstance(self.data_specifier, pyuavcan.transport.ServiceDataSpecifier) and self.remote_node_id is None:
            raise ValueError("Service transfers shall be unicast")

        if isinstance(self.data_specifier, pyuavcan.transport.MessageDataSpecifier) and self.remote_node_id is not None:
            warnings.warn(
                f"Unicast message transfers are an experimental extension of the protocol which "
                f"should not be used in production yet. "
                f"If your application relies on this feature, leave feedback at https://forum.uavcan.org.",
                category=RuntimeWarning,
                stacklevel=-2,
            )

    @property
    def is_broadcast(self) -> bool:
        return self.remote_node_id is None


@dataclasses.dataclass
class SessionStatistics:
    """
    Abstract transport-agnostic session statistics.
    Transport implementations are encouraged to extend this class to add more transport-specific information.
    The statistical counters start from zero when a session is first instantiated.
    """

    transfers: int = 0
    """Successful transfer count."""
    frames: int = 0
    """UAVCAN transport frame count (CAN frames, UDP packets, wireless frames, etc)."""
    payload_bytes: int = 0
    """Successful transfer payload bytes (not including transport metadata or padding)."""
    errors: int = 0
    """Failures of any kind, even if they are also logged using other means, excepting drops."""
    drops: int = 0
    """Frames lost to buffer overruns and expired deadlines."""

    def __eq__(self, other: object) -> bool:
        """
        The statistic comparison operator is defined for any combination of derived classes.
        It compares only those fields that are available in both operands, ignoring unique fields.
        This is useful for testing.
        """
        if isinstance(other, SessionStatistics):
            fds = set(f.name for f in dataclasses.fields(self)) & set(f.name for f in dataclasses.fields(other))
            return all(getattr(self, n) == getattr(other, n) for n in fds)
        return NotImplemented


class Session(abc.ABC):
    """
    Abstract session base class. This is further specialized by input and output.
    Properties should not raise exceptions.
    """

    @property
    @abc.abstractmethod
    def specifier(self) -> SessionSpecifier:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def payload_metadata(self) -> PayloadMetadata:
        raise NotImplementedError

    @abc.abstractmethod
    def sample_statistics(self) -> SessionStatistics:
        """
        Samples and returns the approximated statistics.
        We say "approximated" because implementations are not required to sample the counters atomically,
        although normally they should strive to do so when possible.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        After a session is closed, none of its methods can be used.
        Methods invoked on a closed session should immediately raise :class:`pyuavcan.transport.ResourceClosedError`.
        Subsequent calls to close() will have no effect (no exception either).

        Methods where a task is blocked (such as receive()) at the time of close() will raise a
        :class:`pyuavcan.transport.ResourceClosedError` upon next invocation or sooner.
        Callers of such blocking methods are recommended to avoid usage of large timeouts to facilitate
        faster reaction to transport closure.
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self.specifier, self.payload_metadata)


# noinspection PyAbstractClass
class InputSession(Session):
    """
    Either promiscuous or selective input session.
    The configuration cannot be changed once instantiated.

    Users shall never construct instances themselves;
    instead, the factory method :meth:`pyuavcan.transport.Transport.get_input_session` shall be used.
    """

    @property
    @abc.abstractmethod
    def specifier(self) -> InputSessionSpecifier:
        raise NotImplementedError

    @abc.abstractmethod
    async def receive(self, monotonic_deadline: float) -> typing.Optional[TransferFrom]:
        """
        Attempts to receive the transfer before the deadline [second].
        Returns None if the transfer is not received before the deadline.
        The deadline is compared against :meth:`asyncio.AbstractEventLoop.time`.
        If the deadline is in the past, checks once if there is a transfer and then returns immediately
        without context switching.

        Implementations that use internal queues are recommended to permit the consumer to continue reading
        queued transfers after the instance is closed until the queue is empty.
        In other words, it is recommended to not raise the ResourceClosed exception until
        the instance is closed AND the queue is empty.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def transfer_id_timeout(self) -> float:
        """
        By default, the transfer-ID timeout [second] is initialized with the default value provided in the
        UAVCAN specification.
        It can be overridden using this interface if necessary (rarely is).
        An attempt to assign an invalid timestamp value raises :class:`ValueError`.
        """
        raise NotImplementedError

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        raise NotImplementedError

    @property
    def source_node_id(self) -> typing.Optional[int]:
        """
        Alias for ``.specifier.remote_node_id``.
        For promiscuous sessions this is always None.
        For selective sessions this is the node-ID of the source.
        """
        return self.specifier.remote_node_id


# noinspection PyAbstractClass
class OutputSession(Session):
    """
    Either broadcast or unicast output session.
    The configuration cannot be changed once instantiated.

    Users shall never construct instances themselves;
    instead, the factory method :meth:`pyuavcan.transport.Transport.get_output_session` shall be used.
    """

    @property
    @abc.abstractmethod
    def specifier(self) -> OutputSessionSpecifier:
        raise NotImplementedError

    @abc.abstractmethod
    async def send(self, transfer: Transfer, monotonic_deadline: float) -> bool:
        """
        Sends the transfer; blocks if necessary until the specified deadline [second].
        The deadline value is compared against :meth:`asyncio.AbstractEventLoop.time`.
        Returns when transmission is completed, in which case the return value is True;
        or when the deadline is reached, in which case the return value is False.
        In the case of timeout, a multi-frame transfer may be emitted partially,
        thereby rendering the receiving end unable to process it.
        If the deadline is in the past, the method attempts to send the frames anyway as long as that
        doesn't involve blocking (i.e., task context switching).

        Some transports or media sub-layers may be unable to guarantee transmission strictly before the deadline;
        for example, that may be the case if there is an additional buffering layer under the transport/media
        implementation (e.g., that could be the case with SLCAN-interfaced CAN bus adapters, IEEE 802.15.4 radios,
        and so on, where the data is pushed through an intermediary interface and briefly buffered again before
        being pushed onto the media).
        This is a design limitation imposed by the underlying non-real-time platform that Python runs on;
        it is considered acceptable since PyUAVCAN is designed for soft-real-time applications at most.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def enable_feedback(self, handler: typing.Callable[[Feedback], None]) -> None:
        """
        The output feedback feature makes the transport invoke the specified handler soon after the first
        frame of each transfer originating from this session instance is delivered to the network interface
        or similar underlying logic (not to be confused with delivery to the destination node!).
        This is designed for transmission timestamping, which in turn is necessary for certain protocol features
        such as highly accurate time synchronization.

        The handler is invoked with one argument of type :class:`pyuavcan.transport.Feedback`
        which contains the timing information.
        The transport implementation is allowed to invoke the handler from any context, possibly from another thread.
        The caller should ensure adequate synchronization.
        The actual delay between the emission of the first frame and invocation of the callback is
        implementation-defined, but implementations should strive to minimize it.

        Output feedback is disabled by default. It can be enabled by invoking this method.
        While the feedback is enabled, the performance of the transport in general (not just this session instance)
        may be reduced, possibly resulting in higher input/output latencies and increased CPU load.

        When feedback is already enabled at the time of invocation, this method removes the old callback
        and installs the new one instead.

        Design motivation: We avoid full-transfer loopback such as used in Libuavcan (at least in its old version)
        on purpose because that would make it impossible for us to timestamp outgoing transfers independently
        per transport interface (assuming redundant transports here), since the transport aggregation logic
        would deduplicate redundant received transfers, thus making the valuable timing information unavailable.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def disable_feedback(self) -> None:
        """
        Restores the original state.
        Does nothing if the callback is already disabled.
        """
        raise NotImplementedError

    @property
    def destination_node_id(self) -> typing.Optional[int]:
        """
        Alias for ``.specifier.remote_node_id``.
        For broadcast sessions this is always None.
        For unicast sessions this is the node-ID of the destination.
        """
        return self.specifier.remote_node_id



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import asyncio
import pyuavcan.util
import pyuavcan.dsdl
import pyuavcan.transport
from ._port import OutgoingTransferIDCounter, PortFinalizer, Closable, Port
from ._port import Publisher, PublisherImpl
from ._port import Subscriber, SubscriberImpl
from ._port import Client, ClientImpl
from ._port import Server


MessageClass = typing.TypeVar("MessageClass", bound=pyuavcan.dsdl.CompositeObject)
ServiceClass = typing.TypeVar("ServiceClass", bound=pyuavcan.dsdl.ServiceObject)

FixedPortMessageClass = typing.TypeVar("FixedPortMessageClass", bound=pyuavcan.dsdl.FixedPortCompositeObject)
FixedPortServiceClass = typing.TypeVar("FixedPortServiceClass", bound=pyuavcan.dsdl.FixedPortServiceObject)


_logger = logging.getLogger(__name__)


class Presentation:
    r"""
    This is the presentation layer controller.
    It weaves the fabric of peace and maintains balance even when it looks like the darkest of skies spins above.

    Methods named ``make_*()`` create a new instance upon every invocation. Such instances implement the RAII pattern,
    managing the life cycle of the underlying resource automatically, so the user does not necessarily have to call
    ``close()`` manually, although it is recommended for determinism.

    Methods named ``get_*()`` create a new instance only the first time they are invoked for the
    particular key parameter; the same instance is returned for every subsequent call for the same
    key parameter until it is manually closed by the caller.
    """

    def __init__(self, transport: pyuavcan.transport.Transport) -> None:
        """
        The presentation controller takes ownership of the supplied transport.
        When the presentation instance is closed, its transport is also closed (and so are all its sessions).
        """
        self._transport = transport
        self._closed = False
        self._output_transfer_id_map: typing.Dict[
            pyuavcan.transport.OutputSessionSpecifier, OutgoingTransferIDCounter
        ] = {}
        # For services, the session is the input session.
        self._registry: typing.Dict[
            typing.Tuple[typing.Type[Port[pyuavcan.dsdl.CompositeObject]], pyuavcan.transport.SessionSpecifier],
            Closable,
        ] = {}

    @property
    def output_transfer_id_map(
        self,
    ) -> typing.Dict[pyuavcan.transport.OutputSessionSpecifier, OutgoingTransferIDCounter]:
        """
        This property is designed for very short-lived processes like CLI tools. Most applications will not
        benefit from it and should not use it.

        Access to the output transfer-ID map allows short-running applications
        to store/restore the map to/from a persistent storage that retains data across restarts of the application.
        That may allow applications with very short life cycles (typically under several seconds) to adhere to the
        transfer-ID computation requirements presented in the specification. If the requirement were to be violated,
        then upon restart a process using the same node-ID could be unable to initiate communication using same
        port-ID until the receiving nodes reached the transfer-ID timeout state.

        The typical usage pattern is as follows: Upon launch, check if there is a transfer-ID map stored in a
        predefined location (e.g., a file or a database). If there is, and the storage was last written recently
        (no point restoring a map that is definitely obsolete), load it and commit to this instance by invoking
        :meth:`dict.update` on the object returned by this property. If there isn't, do nothing. When the application
        is finished running (e.g., this could be implemented via :func:`atexit.register`), access the map via this
        property and write it to the predefined storage location atomically. Make sure to shard the location by
        node-ID because nodes that use different node-ID values obviously shall not share their transfer-ID maps.
        Nodes sharing the same node-ID cannot exist on the same transport, but the local system might be running
        nodes under the same node-ID on independent networks concurrently, so this may need to be accounted for.
        """
        return self._output_transfer_id_map

    @property
    def transport(self) -> pyuavcan.transport.Transport:
        """
        Direct reference to the underlying transport instance.
        The presentation layer instance owns its transport.
        """
        return self._transport

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """
        A wrapper for :attr:`pyuavcan.transport.Transport.loop`.
        """
        return self._transport.loop

    # ----------------------------------------  SESSION FACTORY METHODS  ----------------------------------------

    def make_publisher(self, dtype: typing.Type[MessageClass], subject_id: int) -> Publisher[MessageClass]:
        """
        Creates a new publisher instance for the specified subject-ID. All publishers created for a specific
        subject share the same underlying implementation object which is hidden from the user;
        the implementation is reference counted and it is destroyed automatically along with its
        underlying transport level session instance when the last publisher is closed.
        The publisher instance will be closed automatically from the finalizer when garbage collected
        if the user did not bother to do that manually. This logic follows the RAII pattern.

        See :class:`Publisher` for further information about publishers.
        """
        if issubclass(dtype, pyuavcan.dsdl.ServiceObject):
            raise TypeError(f"Not a message type: {dtype}")

        self._raise_if_closed()
        _logger.debug("%s: Constructing new publisher for %r at subject-ID %d", self, dtype, subject_id)

        data_specifier = pyuavcan.transport.MessageDataSpecifier(subject_id)
        session_specifier = pyuavcan.transport.OutputSessionSpecifier(data_specifier, None)
        try:
            impl = self._registry[Publisher, session_specifier]
            assert isinstance(impl, PublisherImpl)
        except LookupError:
            transport_session = self._transport.get_output_session(
                session_specifier, self._make_payload_metadata(dtype)
            )
            transfer_id_counter = self._output_transfer_id_map.setdefault(
                session_specifier, OutgoingTransferIDCounter()
            )
            impl = PublisherImpl(
                dtype=dtype,
                transport_session=transport_session,
                transfer_id_counter=transfer_id_counter,
                finalizer=self._make_finalizer(Publisher, session_specifier),
                loop=self.loop,
            )
            self._registry[Publisher, session_specifier] = impl

        assert isinstance(impl, PublisherImpl)
        return Publisher(impl, self.loop)

    def make_subscriber(
        self, dtype: typing.Type[MessageClass], subject_id: int, queue_capacity: typing.Optional[int] = None
    ) -> Subscriber[MessageClass]:
        """
        Creates a new subscriber instance for the specified subject-ID. All subscribers created for a specific
        subject share the same underlying implementation object which is hidden from the user; the implementation
        is reference counted and it is destroyed automatically along with its underlying transport level session
        instance when the last subscriber is closed. The subscriber instance will be closed automatically from
        the finalizer when garbage collected if the user did not bother to do that manually.
        This logic follows the RAII pattern.

        By default, the size of the input queue is unlimited; the user may provide a positive integer value to override
        this. If the user is not reading the received messages quickly enough and the size of the queue is limited
        (technically, it is always limited at least by the amount of the available memory),
        the queue may become full in which case newer messages will be dropped and the overrun counter
        will be incremented once per dropped message.

        See :class:`Subscriber` for further information about subscribers.
        """
        if issubclass(dtype, pyuavcan.dsdl.ServiceObject):
            raise TypeError(f"Not a message type: {dtype}")

        self._raise_if_closed()
        _logger.debug(
            "%s: Constructing new subscriber for %r at subject-ID %d with queue limit %s",
            self,
            dtype,
            subject_id,
            queue_capacity,
        )

        data_specifier = pyuavcan.transport.MessageDataSpecifier(subject_id)
        session_specifier = pyuavcan.transport.InputSessionSpecifier(data_specifier, None)
        try:
            impl = self._registry[Subscriber, session_specifier]
            assert isinstance(impl, SubscriberImpl)
        except LookupError:
            transport_session = self._transport.get_input_session(session_specifier, self._make_payload_metadata(dtype))
            impl = SubscriberImpl(
                dtype=dtype,
                transport_session=transport_session,
                finalizer=self._make_finalizer(Subscriber, session_specifier),
                loop=self.loop,
            )
            self._registry[Subscriber, session_specifier] = impl

        assert isinstance(impl, SubscriberImpl)
        return Subscriber(impl=impl, loop=self.loop, queue_capacity=queue_capacity)

    def make_client(
        self, dtype: typing.Type[ServiceClass], service_id: int, server_node_id: int
    ) -> Client[ServiceClass]:
        """
        Creates a new client instance for the specified service-ID and the remote server node-ID.
        The number of such instances can be arbitrary.
        For example, different tasks may simultaneously create and use client instances
        invoking the same service on the same server node.

        All clients created with a specific combination of service-ID and server node-ID share the same
        underlying implementation object which is hidden from the user.
        The implementation instance is reference counted and it is destroyed automatically along with its
        underlying transport level session instances when its last client is closed.
        The client instance will be closed automatically from its finalizer when garbage
        collected if the user did not bother to do that manually.
        This logic follows the RAII pattern.

        See :class:`Client` for further information about clients.
        """
        if not issubclass(dtype, pyuavcan.dsdl.ServiceObject):
            raise TypeError(f"Not a service type: {dtype}")

        self._raise_if_closed()
        _logger.debug(
            "%s: Constructing new client for %r at service-ID %d with remote server node-ID %s",
            self,
            dtype,
            service_id,
            server_node_id,
        )

        def transfer_id_modulo_factory() -> int:
            return self._transport.protocol_parameters.transfer_id_modulo

        input_session_specifier = pyuavcan.transport.InputSessionSpecifier(
            pyuavcan.transport.ServiceDataSpecifier(service_id, pyuavcan.transport.ServiceDataSpecifier.Role.RESPONSE),
            server_node_id,
        )
        output_session_specifier = pyuavcan.transport.OutputSessionSpecifier(
            pyuavcan.transport.ServiceDataSpecifier(service_id, pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST),
            server_node_id,
        )
        try:
            impl = self._registry[Client, input_session_specifier]
            assert isinstance(impl, ClientImpl)
        except LookupError:
            output_transport_session = self._transport.get_output_session(
                output_session_specifier, self._make_payload_metadata(dtype.Request)
            )
            input_transport_session = self._transport.get_input_session(
                input_session_specifier, self._make_payload_metadata(dtype.Response)
            )
            transfer_id_counter = self._output_transfer_id_map.setdefault(
                output_session_specifier, OutgoingTransferIDCounter()
            )
            impl = ClientImpl(
                dtype=dtype,
                input_transport_session=input_transport_session,
                output_transport_session=output_transport_session,
                transfer_id_counter=transfer_id_counter,
                transfer_id_modulo_factory=transfer_id_modulo_factory,
                finalizer=self._make_finalizer(Client, input_session_specifier),
                loop=self.loop,
            )
            self._registry[Client, input_session_specifier] = impl

        assert isinstance(impl, ClientImpl)
        return Client(impl=impl, loop=self.loop)

    def get_server(self, dtype: typing.Type[ServiceClass], service_id: int) -> Server[ServiceClass]:
        """
        Returns the server instance for the specified service-ID. If such instance does not exist, it will be
        created. The instance should be used from one task only.

        Observe that unlike other sessions, the server instance is returned as-is without
        any intermediate proxy objects, and this interface does NOT implement the RAII pattern.
        The server instance will not be garbage collected as long as its presentation layer controller exists,
        hence it is the responsibility of the user to close unwanted servers manually.
        However, when the parent presentation layer controller is closed (see :meth:`close`),
        all of its session instances are also closed, servers are no exception, so the application does not
        really have to hunt down every server to terminate a UAVCAN stack properly.

        See :class:`Server` for further information about servers.
        """
        if not issubclass(dtype, pyuavcan.dsdl.ServiceObject):
            raise TypeError(f"Not a service type: {dtype}")

        self._raise_if_closed()
        _logger.debug("%s: Providing server for %r at service-ID %d", self, dtype, service_id)

        def output_transport_session_factory(client_node_id: int) -> pyuavcan.transport.OutputSession:
            _logger.debug("%s: %r has requested a new output session to client node %s", self, impl, client_node_id)
            ds = pyuavcan.transport.ServiceDataSpecifier(
                service_id, pyuavcan.transport.ServiceDataSpecifier.Role.RESPONSE
            )
            return self._transport.get_output_session(
                pyuavcan.transport.OutputSessionSpecifier(ds, client_node_id),
                self._make_payload_metadata(dtype.Response),
            )

        input_session_specifier = pyuavcan.transport.InputSessionSpecifier(
            pyuavcan.transport.ServiceDataSpecifier(service_id, pyuavcan.transport.ServiceDataSpecifier.Role.REQUEST),
            None,
        )
        try:
            impl = self._registry[Server, input_session_specifier]
            assert isinstance(impl, Server)
        except LookupError:
            input_transport_session = self._transport.get_input_session(
                input_session_specifier, self._make_payload_metadata(dtype.Request)
            )
            impl = Server(
                dtype=dtype,
                input_transport_session=input_transport_session,
                output_transport_session_factory=output_transport_session_factory,
                finalizer=self._make_finalizer(Server, input_session_specifier),
                loop=self.loop,
            )
            self._registry[Server, input_session_specifier] = impl

        assert isinstance(impl, Server)
        return impl

    # ----------------------------------------  CONVENIENCE FACTORY METHODS  ----------------------------------------

    def make_publisher_with_fixed_subject_id(
        self, dtype: typing.Type[FixedPortMessageClass]
    ) -> Publisher[FixedPortMessageClass]:
        """
        A wrapper for :meth:`make_publisher` that uses the fixed subject-ID associated with this type.
        Raises a TypeError if the type has no fixed subject-ID.
        """
        return self.make_publisher(dtype=dtype, subject_id=self._get_fixed_port_id(dtype))

    def make_subscriber_with_fixed_subject_id(
        self, dtype: typing.Type[FixedPortMessageClass], queue_capacity: typing.Optional[int] = None
    ) -> Subscriber[FixedPortMessageClass]:
        """
        A wrapper for :meth:`make_subscriber` that uses the fixed subject-ID associated with this type.
        Raises a TypeError if the type has no fixed subject-ID.
        """
        return self.make_subscriber(
            dtype=dtype, subject_id=self._get_fixed_port_id(dtype), queue_capacity=queue_capacity
        )

    def make_client_with_fixed_service_id(
        self, dtype: typing.Type[FixedPortServiceClass], server_node_id: int
    ) -> Client[FixedPortServiceClass]:
        """
        A wrapper for :meth:`make_client` that uses the fixed service-ID associated with this type.
        Raises a TypeError if the type has no fixed service-ID.
        """
        return self.make_client(dtype=dtype, service_id=self._get_fixed_port_id(dtype), server_node_id=server_node_id)

    def get_server_with_fixed_service_id(
        self, dtype: typing.Type[FixedPortServiceClass]
    ) -> Server[FixedPortServiceClass]:
        """
        A wrapper for :meth:`get_server` that uses the fixed service-ID associated with this type.
        Raises a TypeError if the type has no fixed service-ID.
        """
        return self.get_server(dtype=dtype, service_id=self._get_fixed_port_id(dtype))

    # ----------------------------------------  AUXILIARY ENTITIES  ----------------------------------------

    def close(self) -> None:
        """
        Closes the underlying transport instance and all existing session instances.
        I.e., the application is not required to close every session instance explicitly.
        """
        for s in list(self._registry.values()):
            try:
                s.close()
            except Exception as ex:
                _logger.exception("%r.close() could not close session %r: %s", self, s, ex)

        self._closed = True
        self._transport.close()

    def _make_finalizer(
        self,
        session_type: typing.Type[Port[pyuavcan.dsdl.CompositeObject]],
        session_specifier: pyuavcan.transport.SessionSpecifier,
    ) -> PortFinalizer:
        done = False

        def finalizer(transport_sessions: typing.Iterable[pyuavcan.transport.Session]) -> None:
            # So this is rather messy. Observe that a port instance aggregates two distinct resources that
            # must be allocated and deallocated atomically: the local registry entry in this class and the
            # corresponding transport session instance. I don't want to plaster our session objects with locks and
            # container references, so instead I decided to pass the associated resources into the finalizer, which
            # disposes of all resources atomically. This is clearly not very obvious and in the future we should
            # look for a cleaner design. The cleaner design can be retrofitted easily while keeping the API
            # unchanged so this should be easy to fix transparently by bumping only the patch version of the library.
            nonlocal done
            assert not done, "Internal protocol violation: double finalization"
            _logger.debug(
                "%s: Finalizing %s (%s) with transport sessions %s",
                self,
                session_specifier,
                session_type,
                transport_sessions,
            )
            done = True
            try:
                self._registry.pop((session_type, session_specifier))
            except Exception as ex:
                _logger.exception("%s could not remove port for %s: %s", self, session_specifier, ex)

            for ts in transport_sessions:
                try:
                    ts.close()
                except Exception as ex:
                    _logger.exception("%s could not finalize (close) %s: %s", self, ts, ex)

        return finalizer

    @staticmethod
    def _make_payload_metadata(dtype: typing.Type[pyuavcan.dsdl.CompositeObject]) -> pyuavcan.transport.PayloadMetadata:
        extent_bytes = pyuavcan.dsdl.get_extent_bytes(dtype)
        return pyuavcan.transport.PayloadMetadata(extent_bytes=extent_bytes)

    def _raise_if_closed(self) -> None:
        if self._closed:
            raise pyuavcan.transport.ResourceClosedError(repr(self))

    @staticmethod
    def _get_fixed_port_id(dtype: typing.Type[pyuavcan.dsdl.FixedPortObject]) -> int:
        port_id = pyuavcan.dsdl.get_fixed_port_id(dtype)
        if port_id is None:
            raise TypeError(f"{dtype} has no fixed port-ID")
        return port_id

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self,
            self.transport,
            num_publishers=sum(1 for t, _ in self._registry if issubclass(t, Publisher)),
            num_subscribers=sum(1 for t, _ in self._registry if issubclass(t, Subscriber)),
            num_clients=sum(1 for t, _ in self._registry if issubclass(t, Client)),
            num_servers=sum(1 for t, _ in self._registry if issubclass(t, Server)),
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import abc
import typing
import logging
import dataclasses
import pyuavcan.transport


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class RedundantSessionStatistics(pyuavcan.transport.SessionStatistics):
    """
    Aggregate statistics for all inferior sessions in a redundant group.
    This is an atomic immutable sample; it is not updated after construction.
    """

    inferiors: typing.List[pyuavcan.transport.SessionStatistics] = dataclasses.field(default_factory=list)
    """
    The ordering is guaranteed to match that of :attr:`RedundantSession.inferiors`.
    """


class RedundantSession(abc.ABC):
    """
    The base for all redundant session instances.

    A redundant session may be constructed even if the redundant transport itself has no inferiors.
    When a new inferior transport is attached/detached to/from the redundant group,
    dependent session instances are automatically reconfigured, transparently to the user.

    The higher layers of the protocol stack are therefore shielded from any changes made to the stack
    below the redundant transport instance; existing sessions and other instances are never invalidated.
    This guarantee allows one to construct applications whose underlying transport configuration
    can be changed at runtime.
    """

    @property
    @abc.abstractmethod
    def specifier(self) -> pyuavcan.transport.SessionSpecifier:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def inferiors(self) -> typing.Sequence[pyuavcan.transport.Session]:
        """
        Read-only access to the list of inferiors.
        The ordering is guaranteed to match that of :attr:`RedundantTransport.inferiors`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """
        Closes and detaches all inferior sessions.
        If any of the sessions fail to close, an error message will be logged, but no exception will be raised.
        The instance will no longer be usable afterward.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _add_inferior(self, session: pyuavcan.transport.Session) -> None:
        """
        If the new session is already an inferior, this method does nothing.
        If anything goes wrong during the initial setup, the inferior will not be added and
        an appropriate exception will be raised.

        This method is intended to be invoked by the transport class.
        The Python's type system does not allow us to concisely define module-internal APIs.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _close_inferior(self, session_index: int) -> None:
        """
        If the index is out of range, this method does nothing.
        Removal always succeeds regardless of any exceptions raised.

        Like its counterpart, this method is supposed to be invoked by the transport class.
        """
        raise NotImplementedError



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# noinspection PyUnresolvedReferences
r"""
Presentation layer overview
+++++++++++++++++++++++++++

The presentation layer is responsible for serializing and deserializing DSDL objects and for providing a higher-level
object-oriented interface on top of the transport layer.
A typical application should not access this layer directly;
instead, it should rely on the high-level API provided by :mod:`pyuavcan.application`.

The presentation layer uses the term *port* to refer to an instance of publisher, subscriber, service client,
or service server for a specific subject or service (see the inheritance diagram below).

It is possible to create multiple ports that access the same underlying transport layer instance concurrently,
taking care of all related data management and synchronization issues automatically.
This minimizes the logical coupling between different components
of the application that have to rely on the same UAVCAN network resource.
For example, when the application creates more than one subscriber for a given subject, the presentation
layer will distribute received messages into every subscription instance requested by the application.
Likewise, different components of the application may publish messages over the same subject
or invoke the same service on the same remote server node.

Inheritance diagram for the presentation layer is shown below.
Classes named ``*Impl`` are not accessible to the user; their instances are managed automatically by the
presentation layer controller class.
Trivial types may be omitted from the diagram.

.. inheritance-diagram:: pyuavcan.presentation._port._publisher
                         pyuavcan.presentation._port._subscriber
                         pyuavcan.presentation._port._server
                         pyuavcan.presentation._port._client
                         pyuavcan.presentation._port._error
   :parts: 1


Usage example
+++++++++++++

.. attention::
    A typical application should not instantiate presentation-layer entities directly;
    instead, use the higher-level API provided by :mod:`pyuavcan.application`.

The main entity of the presentation layer is the class :class:`pyuavcan.presentation.Presentation`;
the following demo shows how it can be used.
This example is based on a simple loopback transport that does not interact with the outside world
(it doesn't perform IO with the OS), which makes it well-suited for demo needs.

..  doctest::
    :hide:

    >>> import tests
    >>> _ = tests.dsdl.compile()

>>> import uavcan.node, uavcan.diagnostic         # Import what we need from DSDL-generated packages.
>>> import pyuavcan.transport.loopback            # Import the demo transport implementation.
>>> transport = pyuavcan.transport.loopback.LoopbackTransport(None)  # Use your real transport instead.
>>> presentation = pyuavcan.presentation.Presentation(transport)

Having prepared a presentation layer controller, we can create *ports*.
They are the main points of network access for the application.
Let's start with a publisher and a subscriber:

>>> pub_record = presentation.make_publisher_with_fixed_subject_id(uavcan.diagnostic.Record_1_1)
>>> sub_record = presentation.make_subscriber_with_fixed_subject_id(uavcan.diagnostic.Record_1_1)

Publish a message and receive it also (the loopback transport just returns all outgoing transfers back):

>>> import asyncio
>>> run_until_complete = asyncio.get_event_loop().run_until_complete
>>> record = uavcan.diagnostic.Record_1_1(
...     severity=uavcan.diagnostic.Severity_1_0(uavcan.diagnostic.Severity_1_0.INFO),
...     text='Neither man nor animal can be influenced by anything but suggestion.')
>>> run_until_complete(pub_record.publish(record))  # publish() returns False on timeout.
True
>>> message, metadata = run_until_complete(sub_record.receive_for(timeout=0.5))
>>> message.text.tobytes().decode()  # Calling .tobytes().decode() won't be needed when DSDL supports strings natively.
'Neither man nor animal can be influenced by anything but suggestion.'
>>> metadata.transfer_id, metadata.source_node_id, metadata.timestamp
(0, None, Timestamp(system_ns=..., monotonic_ns=...))

We can use custom subject-ID with any data type, even if there is a fixed subject-ID provided
(the background is explained in Specification, please read it).
Here is an example; we also show here that when a receive call times out, it returns None:

>>> sub_record_custom = presentation.make_subscriber(uavcan.diagnostic.Record_1_1, subject_id=2345)
>>> run_until_complete(sub_record_custom.receive_for(timeout=0.5))  # Times out and returns None.

You can see above that the node-ID of the received transfer metadata is None,
that's because it is actually an anonymous transfer, and it is so because our node is an anonymous node;
i.e., it doesn't have a node-ID.

>>> presentation.transport.local_node_id is None    # Yup, it's anonymous.
True

Next we're going to create a service.
Services can't be used with anonymous nodes (which is natural -- how do you send a unicast transfer
to an anonymous node?), so we'll have to create a new transport with a node-ID of its own.

>>> transport = pyuavcan.transport.loopback.LoopbackTransport(1234)  # The range of valid values is transport-dependent.
>>> presentation = pyuavcan.presentation.Presentation(transport)  # Start anew, this time not anonymous.
>>> presentation.transport.local_node_id
1234

Generally, anonymous nodes are useful in two cases:

1. You only need to listen and you know that you are not going to emit any transfers
   (no point tinkering with node-ID if you're not going to use it anyway).

2. You need to allocate a node-ID using the plug-and-play autoconfiguration protocol.
   In this case, you would normally create a transport, run the PnP allocation procedure to obtain a node-ID value
   from the PnP allocator, and then replace your transport instance with a new one (similar to what we just did here)
   initialized with the node-ID value provided by the PnP allocator.


Having configured the node-ID, let's set up a service and invoke it:

>>> async def on_request(request: uavcan.node.ExecuteCommand_1_1.Request,
...                      metadata: pyuavcan.presentation.ServiceRequestMetadata) \
...         -> uavcan.node.ExecuteCommand_1_1.Response:
...     print(f'Received command {request.command} from node {metadata.client_node_id}')
...     return uavcan.node.ExecuteCommand_1_1.Response(uavcan.node.ExecuteCommand_1_1.Response.STATUS_BAD_COMMAND)
>>> srv_exec_command = presentation.get_server_with_fixed_service_id(uavcan.node.ExecuteCommand_1_1)
>>> srv_exec_command.serve_in_background(on_request)
>>> client_exec_command = presentation.make_client_with_fixed_service_id(uavcan.node.ExecuteCommand_1_1,
...                                                                      server_node_id=1234)
>>> request_object = uavcan.node.ExecuteCommand_1_1.Request(
...     uavcan.node.ExecuteCommand_1_1.Request.COMMAND_BEGIN_SOFTWARE_UPDATE,
...     '/path/to/the/firmware/image.bin')
>>> received_response, response_transfer = run_until_complete(client_exec_command.call(request_object))
Received command 65533 from node 1234
>>> received_response
uavcan.node.ExecuteCommand.Response.1.1(status=3)

Methods that receive data from the network return None on timeout.
For example, here we create a client for a nonexistent service; the call times out and returns None:

>>> bad_client = presentation.make_client(uavcan.node.ExecuteCommand_1_1,
...                                       service_id=234,       # There is no such service.
...                                       server_node_id=321)   # There is no such server.
>>> bad_client.response_timeout = 0.1                           # Override the default.
>>> bad_client.priority = pyuavcan.transport.Priority.HIGH      # Override the default.
>>> run_until_complete(bad_client.call(request_object))         # Times out and returns None.
"""

from ._presentation import Presentation as Presentation

from ._port import Publisher as Publisher
from ._port import Subscriber as Subscriber
from ._port import Client as Client
from ._port import Server as Server

from ._port import SubscriberStatistics as SubscriberStatistics
from ._port import ClientStatistics as ClientStatistics
from ._port import ServerStatistics as ServerStatistics
from ._port import ServiceRequestMetadata as ServiceRequestMetadata
from ._port import ServiceRequestHandler as ServiceRequestHandler

from ._port import Port as Port
from ._port import MessagePort as MessagePort
from ._port import ServicePort as ServicePort

from ._port import OutgoingTransferIDCounter as OutgoingTransferIDCounter
from ._port import PortClosedError as PortClosedError
from ._port import RequestTransferIDVariabilityExhaustedError as RequestTransferIDVariabilityExhaustedError
from ._port import DEFAULT_PRIORITY as DEFAULT_PRIORITY
from ._port import DEFAULT_SERVICE_REQUEST_TIMEOUT as DEFAULT_SERVICE_REQUEST_TIMEOUT



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import logging
import pyuavcan
from pyuavcan.transport import ServiceDataSpecifier
from .._frame import SerialFrame
from ._base import SerialSession


#: Returns the transmission timestamp.
SendHandler = typing.Callable[
    [typing.List[SerialFrame], float], typing.Awaitable[typing.Optional[pyuavcan.transport.Timestamp]]
]

_logger = logging.getLogger(__name__)


class SerialFeedback(pyuavcan.transport.Feedback):
    def __init__(
        self,
        original_transfer_timestamp: pyuavcan.transport.Timestamp,
        first_frame_transmission_timestamp: pyuavcan.transport.Timestamp,
    ):
        self._original_transfer_timestamp = original_transfer_timestamp
        self._first_frame_transmission_timestamp = first_frame_transmission_timestamp

    @property
    def original_transfer_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._original_transfer_timestamp

    @property
    def first_frame_transmission_timestamp(self) -> pyuavcan.transport.Timestamp:
        return self._first_frame_transmission_timestamp


class SerialOutputSession(SerialSession, pyuavcan.transport.OutputSession):
    def __init__(
        self,
        specifier: pyuavcan.transport.OutputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        mtu: int,
        local_node_id: typing.Optional[int],
        send_handler: SendHandler,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly.
        Instead, use the factory method :meth:`pyuavcan.transport.serial.SerialTransport.get_output_session`.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._mtu = int(mtu)
        self._local_node_id = local_node_id
        self._send_handler = send_handler
        self._feedback_handler: typing.Optional[typing.Callable[[pyuavcan.transport.Feedback], None]] = None
        self._statistics = pyuavcan.transport.SessionStatistics()
        if self._local_node_id is None and isinstance(self._specifier.data_specifier, ServiceDataSpecifier):
            raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                f"Anonymous nodes cannot emit service transfers. Session specifier: {self._specifier}"
            )
        assert isinstance(self._local_node_id, int) or self._local_node_id is None
        assert callable(send_handler)
        assert (
            specifier.remote_node_id is not None if isinstance(specifier.data_specifier, ServiceDataSpecifier) else True
        ), "Internal protocol violation: cannot broadcast a service transfer"

        super().__init__(finalizer)

    async def send(self, transfer: pyuavcan.transport.Transfer, monotonic_deadline: float) -> bool:
        self._raise_if_closed()

        def construct_frame(index: int, end_of_transfer: bool, payload: memoryview) -> SerialFrame:
            if not end_of_transfer and self._local_node_id is None:
                raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                    f"Anonymous nodes cannot emit multi-frame transfers. Session specifier: {self._specifier}"
                )
            return SerialFrame(
                priority=transfer.priority,
                transfer_id=transfer.transfer_id,
                index=index,
                end_of_transfer=end_of_transfer,
                payload=payload,
                source_node_id=self._local_node_id,
                destination_node_id=self._specifier.remote_node_id,
                data_specifier=self._specifier.data_specifier,
            )

        frames = list(
            pyuavcan.transport.commons.high_overhead_transport.serialize_transfer(
                transfer.fragmented_payload, self._mtu, construct_frame
            )
        )
        _logger.debug("%s: Sending transfer: %s; current stats: %s", self, transfer, self._statistics)
        try:
            tx_timestamp = await self._send_handler(frames, monotonic_deadline)
        except Exception:
            self._statistics.errors += 1
            raise

        if tx_timestamp is not None:
            self._statistics.transfers += 1
            self._statistics.frames += len(frames)
            self._statistics.payload_bytes += sum(map(len, transfer.fragmented_payload))
            if self._feedback_handler is not None:
                try:
                    self._feedback_handler(SerialFeedback(transfer.timestamp, tx_timestamp))
                except Exception as ex:  # pragma: no cover
                    _logger.exception(
                        "Unhandled exception in the output session feedback handler %s: %s", self._feedback_handler, ex
                    )
            return True
        self._statistics.drops += len(frames)
        return False

    def enable_feedback(self, handler: typing.Callable[[pyuavcan.transport.Feedback], None]) -> None:
        self._feedback_handler = handler

    def disable_feedback(self) -> None:
        self._feedback_handler = None

    @property
    def specifier(self) -> pyuavcan.transport.OutputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> pyuavcan.transport.SessionStatistics:
        return copy.copy(self._statistics)

    def close(self) -> None:  # pylint: disable=useless-super-delegation
        super().close()


def _unittest_output_session() -> None:
    import asyncio
    from pytest import raises, approx
    from pyuavcan.transport import OutputSessionSpecifier, MessageDataSpecifier, Priority
    from pyuavcan.transport import PayloadMetadata, SessionStatistics, Timestamp, Feedback, Transfer

    ts = Timestamp.now()
    loop = asyncio.get_event_loop()
    run_until_complete = loop.run_until_complete

    tx_timestamp: typing.Optional[Timestamp] = Timestamp.now()
    tx_exception: typing.Optional[Exception] = None
    last_sent_frames: typing.List[SerialFrame] = []
    last_monotonic_deadline = 0.0
    finalized = False

    async def do_send(frames: typing.Sequence[SerialFrame], monotonic_deadline: float) -> typing.Optional[Timestamp]:
        nonlocal last_sent_frames
        nonlocal last_monotonic_deadline
        last_sent_frames = list(frames)
        last_monotonic_deadline = monotonic_deadline
        if tx_exception:
            raise tx_exception
        return tx_timestamp

    def do_finalize() -> None:
        nonlocal finalized
        finalized = True

    with raises(pyuavcan.transport.OperationNotDefinedForAnonymousNodeError):
        SerialOutputSession(
            specifier=OutputSessionSpecifier(ServiceDataSpecifier(321, ServiceDataSpecifier.Role.REQUEST), 1111),
            payload_metadata=PayloadMetadata(1024),
            mtu=10,
            local_node_id=None,
            send_handler=do_send,
            finalizer=do_finalize,
        )

    sos = SerialOutputSession(
        specifier=OutputSessionSpecifier(MessageDataSpecifier(3210), None),
        payload_metadata=PayloadMetadata(1024),
        mtu=11,
        local_node_id=None,
        send_handler=do_send,
        finalizer=do_finalize,
    )

    assert sos.specifier == OutputSessionSpecifier(MessageDataSpecifier(3210), None)
    assert sos.destination_node_id is None
    assert sos.payload_metadata == PayloadMetadata(1024)
    assert sos.sample_statistics() == SessionStatistics()

    assert run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.NOMINAL,
                transfer_id=12340,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            999999999.999,
        )
    )
    assert last_monotonic_deadline == approx(999999999.999)
    assert len(last_sent_frames) == 1

    with raises(pyuavcan.transport.OperationNotDefinedForAnonymousNodeError):
        run_until_complete(
            sos.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.NOMINAL,
                    transfer_id=12340,
                    fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three four five")],
                ),
                loop.time() + 10.0,
            )
        )

    last_feedback: typing.Optional[Feedback] = None

    def feedback_handler(feedback: Feedback) -> None:
        nonlocal last_feedback
        last_feedback = feedback

    sos.enable_feedback(feedback_handler)

    assert last_feedback is None
    assert run_until_complete(
        sos.send(
            Transfer(timestamp=ts, priority=Priority.NOMINAL, transfer_id=12340, fragmented_payload=[]), 999999999.999
        )
    )
    assert last_monotonic_deadline == approx(999999999.999)
    assert len(last_sent_frames) == 1
    assert last_feedback is not None
    assert last_feedback.original_transfer_timestamp == ts
    assert last_feedback.first_frame_transmission_timestamp == tx_timestamp

    sos.disable_feedback()
    sos.disable_feedback()  # Idempotency check

    assert sos.sample_statistics() == SessionStatistics(transfers=2, frames=2, payload_bytes=11, errors=0, drops=0)

    assert not finalized
    sos.close()
    assert finalized
    finalized = False

    sos = SerialOutputSession(
        specifier=OutputSessionSpecifier(ServiceDataSpecifier(321, ServiceDataSpecifier.Role.REQUEST), 2222),
        payload_metadata=PayloadMetadata(1024),
        mtu=10,
        local_node_id=1234,
        send_handler=do_send,
        finalizer=do_finalize,
    )

    # Induced failure
    tx_timestamp = None
    assert not run_until_complete(
        sos.send(
            Transfer(
                timestamp=ts,
                priority=Priority.NOMINAL,
                transfer_id=12340,
                fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
            ),
            999999999.999,
        )
    )
    assert last_monotonic_deadline == approx(999999999.999)
    assert len(last_sent_frames) == 2

    assert sos.sample_statistics() == SessionStatistics(transfers=0, frames=0, payload_bytes=0, errors=0, drops=2)

    tx_exception = RuntimeError()
    with raises(RuntimeError):
        _ = run_until_complete(
            sos.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.NOMINAL,
                    transfer_id=12340,
                    fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
                ),
                loop.time() + 10.0,
            )
        )

    assert sos.sample_statistics() == SessionStatistics(transfers=0, frames=0, payload_bytes=0, errors=1, drops=2)

    assert not finalized
    sos.close()
    assert finalized
    sos.close()  # Idempotency

    with raises(pyuavcan.transport.ResourceClosedError):
        run_until_complete(
            sos.send(
                Transfer(
                    timestamp=ts,
                    priority=Priority.NOMINAL,
                    transfer_id=12340,
                    fragmented_payload=[memoryview(b"one"), memoryview(b"two"), memoryview(b"three")],
                ),
                loop.time() + 10.0,
            )
        )



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import dataclasses
import pyuavcan
import pyuavcan.transport.serial
from pyuavcan.transport import Trace, TransferTrace, Capture, AlienSessionSpecifier, AlienTransferMetadata
from pyuavcan.transport import AlienTransfer, TransferFrom, Timestamp
from pyuavcan.transport.commons.high_overhead_transport import AlienTransferReassembler, TransferReassembler
from ._frame import SerialFrame
from ._stream_parser import StreamParser


_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class SerialCapture(pyuavcan.transport.Capture):
    """
    Since UAVCAN/serial operates on top of unstructured L1 data links, there is no native concept of framing.
    Therefore, the capture type defines only the timestamp, a raw chunk of bytes, and the direction (RX/TX).

    When capturing data from a live interface, it is guaranteed by this library that each capture will contain
    AT MOST one frame along with the delimiter bytes (at least the last byte of the fragment is zero).
    When reading data from a file, it is trivial to split the data into frames by looking for the frame separators,
    which are simply zero bytes.
    """

    fragment: memoryview

    own: bool
    """
    True if the captured fragment was sent by the local transport instance.
    False if it was received from the port.
    """

    def __repr__(self) -> str:
        """
        Captures that contain large fragments are truncated and appended with an ellipsis.
        """
        limit = 64
        if len(self.fragment) > limit:
            fragment = bytes(self.fragment[:limit]).hex() + f"...<+{len(self.fragment) - limit}B>..."
        else:
            fragment = bytes(self.fragment).hex()
        direction = "tx" if self.own else "rx"
        return pyuavcan.util.repr_attributes(self, direction, fragment)

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.serial.SerialTransport]:
        return pyuavcan.transport.serial.SerialTransport


@dataclasses.dataclass(frozen=True)
class SerialErrorTrace(pyuavcan.transport.ErrorTrace):
    error: TransferReassembler.Error


@dataclasses.dataclass(frozen=True)
class SerialOutOfBandTrace(pyuavcan.transport.ErrorTrace):
    """
    Out-of-band data or a malformed frame received. See :class:`pyuavcan.serial.StreamParser`.
    """

    data: memoryview


class SerialTracer(pyuavcan.transport.Tracer):
    """
    This tracer does not differentiate between input and output traces,
    but it keeps separate parsers for input and output captures such that there is no RX/TX state conflict.
    If necessary, the user can distinguish RX/TX traces by checking :attr:`SerialCapture.direction`
    before invoking :meth:`update`.

    Return types from :meth:`update`:

    - :class:`pyuavcan.transport.TransferTrace`
    - :class:`SerialErrorTrace`
    - :class:`SerialOutOfBandTrace`
    """

    _MTU = 2 ** 32
    """Effectively unlimited."""

    def __init__(self) -> None:
        self._parsers = [
            StreamParser(self._on_parsed, self._MTU),
            StreamParser(self._on_parsed, self._MTU),
        ]
        self._parser_output: typing.Optional[typing.Tuple[Timestamp, typing.Union[SerialFrame, memoryview]]] = None
        self._sessions: typing.Dict[AlienSessionSpecifier, _AlienSession] = {}

    def update(self, cap: Capture) -> typing.Optional[Trace]:
        """
        If the capture encapsulates more than one serialized frame, a :class:`ValueError` will be raised.
        To avoid this, always ensure that the captured fragments are split on the frame delimiters
        (which are simply zero bytes).
        Captures provided by PyUAVCAN are always fragmented correctly, but you may need to implement fragmentation
        manually when reading data from an external file.
        """
        if not isinstance(cap, SerialCapture):
            return None

        self._parsers[cap.own].process_next_chunk(cap.fragment, cap.timestamp)
        if self._parser_output is None:
            return None

        timestamp, item = self._parser_output
        self._parser_output = None
        if isinstance(item, memoryview):
            return SerialOutOfBandTrace(timestamp, item)

        if isinstance(item, SerialFrame):
            spec = AlienSessionSpecifier(
                source_node_id=item.source_node_id,
                destination_node_id=item.destination_node_id,
                data_specifier=item.data_specifier,
            )
            return self._get_session(spec).update(timestamp, item)

        assert False

    def _get_session(self, specifier: AlienSessionSpecifier) -> _AlienSession:
        try:
            return self._sessions[specifier]
        except KeyError:
            self._sessions[specifier] = _AlienSession(specifier)
        return self._sessions[specifier]

    def _on_parsed(self, timestamp: Timestamp, data: memoryview, frame: typing.Optional[SerialFrame]) -> None:
        _logger.debug(
            "Stream parser output (conflict: %s): %s <%d bytes> %s",
            bool(self._parser_output),
            timestamp,
            len(data),
            frame,
        )
        if self._parser_output is None:
            self._parser_output = timestamp, (data if frame is None else frame)
        else:
            self._parser_output = None
            raise ValueError(
                f"The supplied serial capture object contains more than one serialized entity. "
                f"Such arrangement cannot be processed correctly by this implementation. "
                f"Please update the caller code to always fragment the input byte stream at the frame delimiters, "
                f"which are simply zero bytes. "
                f"The timestamp of the offending capture is {timestamp}."
            )


class _AlienSession:
    def __init__(self, specifier: AlienSessionSpecifier) -> None:
        self._specifier = specifier
        src = specifier.source_node_id
        self._reassembler = AlienTransferReassembler(src) if src is not None else None

    def update(self, timestamp: Timestamp, frame: SerialFrame) -> typing.Optional[Trace]:
        reasm = self._reassembler
        tid_timeout = reasm.transfer_id_timeout if reasm is not None else 0.0

        tr: typing.Union[TransferFrom, TransferReassembler.Error, None]
        if reasm is not None:
            tr = reasm.process_frame(timestamp, frame)
        else:
            tr = TransferReassembler.construct_anonymous_transfer(timestamp, frame)

        if isinstance(tr, TransferReassembler.Error):
            return SerialErrorTrace(timestamp=timestamp, error=tr)

        if isinstance(tr, TransferFrom):
            meta = AlienTransferMetadata(tr.priority, tr.transfer_id, self._specifier)
            return TransferTrace(timestamp, AlienTransfer(meta, tr.fragmented_payload), tid_timeout)

        assert tr is None
        return None


# ----------------------------------------  TESTS GO BELOW THIS LINE  ----------------------------------------


def _unittest_serial_tracer() -> None:
    from pytest import raises, approx
    from pyuavcan.transport import Priority, MessageDataSpecifier
    from pyuavcan.transport.serial import SerialTransport

    tr = SerialTransport.make_tracer()
    ts = Timestamp.now()

    def tx(x: typing.Union[bytes, bytearray, memoryview]) -> typing.Optional[Trace]:
        return tr.update(SerialCapture(ts, memoryview(x), own=True))

    def rx(x: typing.Union[bytes, bytearray, memoryview]) -> typing.Optional[Trace]:
        return tr.update(SerialCapture(ts, memoryview(x), own=False))

    buf = SerialFrame(
        priority=Priority.SLOW,
        transfer_id=1234567890,
        index=0,
        end_of_transfer=True,
        payload=memoryview(b"abc"),
        source_node_id=1111,
        destination_node_id=None,
        data_specifier=MessageDataSpecifier(6666),
    ).compile_into(bytearray(100))
    head, tail = buf[:10], buf[10:]

    assert None is tx(head)  # Semi-complete.

    trace = tx(head)  # Double-head invalidates the previous one.
    assert isinstance(trace, SerialOutOfBandTrace)
    assert trace.timestamp == ts
    assert trace.data.tobytes().strip(b"\0") == head.tobytes().strip(b"\0")

    trace = tx(tail)
    assert isinstance(trace, TransferTrace)
    assert trace.timestamp == ts
    assert trace.transfer_id_timeout == approx(2.0)  # Initial value.
    assert trace.transfer.metadata.transfer_id == 1234567890
    assert trace.transfer.metadata.priority == Priority.SLOW
    assert trace.transfer.metadata.session_specifier.source_node_id == 1111
    assert trace.transfer.metadata.session_specifier.destination_node_id is None
    assert trace.transfer.metadata.session_specifier.data_specifier == MessageDataSpecifier(6666)
    assert trace.transfer.fragmented_payload == [memoryview(b"abc")]

    buf = SerialFrame(
        priority=Priority.SLOW,
        transfer_id=1234567890,
        index=0,
        end_of_transfer=True,
        payload=memoryview(b"abc"),
        source_node_id=None,
        destination_node_id=None,
        data_specifier=MessageDataSpecifier(6666),
    ).compile_into(bytearray(100))

    trace = rx(buf)
    assert isinstance(trace, TransferTrace)
    assert trace.timestamp == ts
    assert trace.transfer.metadata.transfer_id == 1234567890
    assert trace.transfer.metadata.session_specifier.source_node_id is None
    assert trace.transfer.metadata.session_specifier.destination_node_id is None

    assert None is tr.update(pyuavcan.transport.Capture(ts))  # Wrong type, ignore.

    trace = tx(
        SerialFrame(
            priority=Priority.SLOW,
            transfer_id=1234567890,
            index=0,
            end_of_transfer=False,
            payload=memoryview(bytes(range(256))),
            source_node_id=3333,
            destination_node_id=None,
            data_specifier=MessageDataSpecifier(6666),
        ).compile_into(bytearray(10_000))
    )
    assert trace is None
    trace = tx(
        SerialFrame(
            priority=Priority.SLOW,
            transfer_id=1234567890,
            index=1,
            end_of_transfer=True,
            payload=memoryview(bytes(range(256))),
            source_node_id=3333,
            destination_node_id=None,
            data_specifier=MessageDataSpecifier(6666),
        ).compile_into(bytearray(10_000))
    )
    assert isinstance(trace, SerialErrorTrace)
    assert trace.error == TransferReassembler.Error.MULTIFRAME_INTEGRITY_ERROR

    with raises(ValueError, match=".*delimiters.*"):
        rx(b"".join([buf, buf]))



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import gzip
import typing
import pickle
import base64
import logging
import importlib

import pydsdl

from . import _serialized_representation


_logger = logging.getLogger(__name__)


class CompositeObject(abc.ABC):  # Members are surrounded with underscores to avoid collisions with DSDL attributes.
    """
    This is the base class for all Python classes generated from DSDL definitions.
    It does not have any public members.
    """

    _MODEL_: pydsdl.CompositeType
    """Type definition as provided by PyDSDL."""

    _EXTENT_BYTES_: int
    """Defined in generated classes."""

    @abc.abstractmethod
    def _serialize_(self, _ser_: _serialized_representation.Serializer) -> None:
        """
        Auto-generated serialization method.
        Appends the serialized representation of its object to the supplied Serializer instance.
        This is not a part of the API.
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _deserialize_(_des_: _serialized_representation.Deserializer) -> CompositeObject:
        """
        Auto-generated deserialization method. Consumes (some) data from the supplied Deserializer instance.
        Raises a Deserializer.FormatError if the supplied serialized representation is invalid.
        Always returns a valid object unless an exception is raised.
        This is not a part of the API.
        """
        raise NotImplementedError

    @staticmethod
    def _restore_constant_(encoded_string: str) -> object:
        """Recovers a pickled gzipped constant object from base85 string representation."""
        out = pickle.loads(gzip.decompress(base64.b85decode(encoded_string)))
        assert isinstance(out, object)
        return out

    # These typing hints are provided here for use in the generated classes. They are obviously not part of the API.
    _SerializerTypeVar_ = typing.TypeVar("_SerializerTypeVar_", bound=_serialized_representation.Serializer)
    _DeserializerTypeVar_ = typing.TypeVar("_DeserializerTypeVar_", bound=_serialized_representation.Deserializer)


class ServiceObject(CompositeObject):
    """
    This is the base class for all Python classes generated from DSDL service type definitions.
    Observe that it inherits from the composite object class, just like the nested types Request and Response.
    """

    Request: typing.Type[CompositeObject]
    """
    Nested request type. Inherits from :class:`CompositeObject`.
    The base class provides a stub which is overridden in generated classes.
    """

    Response: typing.Type[CompositeObject]
    """
    Nested response type. Inherits from :class:`CompositeObject`.
    The base class provides a stub which is overridden in generated classes.
    """

    _EXTENT_BYTES_ = 0

    def _serialize_(self, _ser_: _serialized_representation.Serializer) -> None:
        raise TypeError(f"Service type {type(self).__name__} cannot be serialized")

    @staticmethod
    def _deserialize_(_des_: _serialized_representation.Deserializer) -> CompositeObject:
        raise TypeError("Service types cannot be deserialized")


class FixedPortObject(abc.ABC):
    """
    This is the base class for all Python classes generated from DSDL types that have a fixed port identifier.
    """

    _FIXED_PORT_ID_: int


class FixedPortCompositeObject(CompositeObject, FixedPortObject):
    @abc.abstractmethod
    def _serialize_(self, _ser_: _serialized_representation.Serializer) -> None:
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _deserialize_(_des_: _serialized_representation.Deserializer) -> CompositeObject:
        raise NotImplementedError


class FixedPortServiceObject(ServiceObject, FixedPortObject):
    pass


CompositeObjectTypeVar = typing.TypeVar("CompositeObjectTypeVar", bound=CompositeObject)


def serialize(obj: CompositeObject) -> typing.Iterable[memoryview]:
    """
    Constructs a serialized representation of the provided top-level object.
    The resulting serialized representation is padded to one byte in accordance with the UAVCAN specification.
    The constructed serialized representation is returned as a sequence of byte-aligned fragments which must be
    concatenated in order to obtain the final representation.
    The objective of this model is to avoid copying data into a temporary buffer when possible.
    Each yielded fragment is of type :class:`memoryview` pointing to raw unsigned bytes.
    It is guaranteed that at least one fragment is always returned (which may be empty).
    """
    # TODO: update the Serializer class to emit an iterable of fragments.
    ser = _serialized_representation.Serializer.new(obj._EXTENT_BYTES_)  # pylint: disable=protected-access
    obj._serialize_(ser)  # pylint: disable=protected-access
    yield ser.buffer.data


def deserialize(
    dtype: typing.Type[CompositeObjectTypeVar], fragmented_serialized_representation: typing.Sequence[memoryview]
) -> typing.Optional[CompositeObjectTypeVar]:
    """
    Constructs an instance of the supplied DSDL-generated data type from its serialized representation.
    Returns None if the provided serialized representation is invalid.

    This function will never raise an exception for invalid input data; the only possible outcome of an invalid data
    being supplied is None at the output. A raised exception can only indicate an error in the deserialization logic.

    .. important:: The constructed object may contain arrays referencing the memory allocated for the serialized
        representation. Therefore, in order to avoid unintended data corruption, the caller should destroy all
        references to the serialized representation immediately after the invocation.

    .. important:: The supplied fragments of the serialized representation should be writeable.
        If they are not, some of the array-typed fields of the constructed object may be read-only.
    """
    deserializer = _serialized_representation.Deserializer.new(fragmented_serialized_representation)
    try:
        return dtype._deserialize_(deserializer)  # type: ignore    # pylint: disable=protected-access
    except _serialized_representation.Deserializer.FormatError:
        _logger.info("Invalid serialized representation of %s: %s", get_model(dtype), deserializer, exc_info=True)
        return None


def get_model(class_or_instance: typing.Union[typing.Type[CompositeObject], CompositeObject]) -> pydsdl.CompositeType:
    """
    Obtains a PyDSDL model of the supplied DSDL-generated class or its instance.
    This is the inverse of :func:`get_class`.
    """
    out = class_or_instance._MODEL_  # pylint: disable=protected-access
    assert isinstance(out, pydsdl.CompositeType)
    return out


def get_class(model: pydsdl.CompositeType) -> typing.Type[CompositeObject]:
    """
    Returns a generated native class implementing the specified DSDL type represented by its PyDSDL model object.
    Promotes the model to delimited type automatically if necessary.
    This is the inverse of :func:`get_model`.

    :raises:
        - :class:`ImportError` if the generated package or subpackage cannot be found.

        - :class:`AttributeError` if the package is found but it does not contain the requested type.

        - :class:`TypeError` if the requested type is found, but its model does not match the input argument.
          This error may occur if the DSDL source has changed since the type was generated.
          To fix this, regenerate the package and make sure that all components of the application use identical
          or compatible DSDL source files.
    """

    def do_import(name_components: typing.List[str]) -> typing.Any:
        mod = None
        for comp in name_components:
            name = (mod.__name__ + "." + comp) if mod else comp  # type: ignore
            try:
                mod = importlib.import_module(name)
            except ImportError:  # We seem to have hit a reserved word; try with an underscore.
                mod = importlib.import_module(name + "_")
        return mod

    if model.has_parent_service:  # uavcan.node.GetInfo.Request --> uavcan.node.GetInfo then Request
        parent_name, child_name = model.name_components[-2:]
        mod = do_import(model.name_components[:-2])
        out = getattr(mod, f"{parent_name}_{model.version.major}_{model.version.minor}")
        assert issubclass(out, ServiceObject)
        out = getattr(out, child_name)
    else:
        mod = do_import(model.name_components[:-1])
        out = getattr(mod, f"{model.short_name}_{model.version.major}_{model.version.minor}")

    out_model = get_model(out)
    if out_model.inner_type != model.inner_type:
        raise TypeError(
            f"The class has been generated using an incompatible DSDL definition. "
            f"Requested model: {model} defined in {model.source_file_path}. "
            f"Model found in the class: {out_model} defined in {out_model.source_file_path}."
        )

    assert str(get_model(out)) == str(model)
    assert isinstance(out, type)
    assert issubclass(out, CompositeObject)
    return out


def get_extent_bytes(class_or_instance: typing.Union[typing.Type[CompositeObject], CompositeObject]) -> int:
    return int(class_or_instance._EXTENT_BYTES_)  # pylint: disable=protected-access


def get_fixed_port_id(
    class_or_instance: typing.Union[typing.Type[FixedPortObject], FixedPortObject]
) -> typing.Optional[int]:
    """
    Returns None if the supplied type has no fixed port-ID.
    """
    try:
        out = int(class_or_instance._FIXED_PORT_ID_)  # pylint: disable=protected-access
    except (TypeError, AttributeError):
        return None
    else:
        if (isinstance(class_or_instance, type) and issubclass(class_or_instance, CompositeObject)) or isinstance(
            class_or_instance, CompositeObject
        ):  # pragma: no branch
            assert out == get_model(class_or_instance).fixed_port_id
        return out


def get_attribute(obj: typing.Union[CompositeObject, typing.Type[CompositeObject]], name: str) -> typing.Any:
    """
    DSDL type attributes whose names can't be represented in Python (such as ``def`` or ``type``)
    are suffixed with an underscore.
    This function allows the caller to read arbitrary attributes referring to them by their original
    DSDL names, e.g., ``def`` instead of ``def_``.

    This function behaves like :func:`getattr` if the attribute does not exist.
    """
    try:
        return getattr(obj, name)
    except AttributeError:
        return getattr(obj, name + "_")


def set_attribute(obj: CompositeObject, name: str, value: typing.Any) -> None:
    """
    DSDL type attributes whose names can't be represented in Python (such as ``def`` or ``type``)
    are suffixed with an underscore.
    This function allows the caller to assign arbitrary attributes referring to them by their original DSDL names,
    e.g., ``def`` instead of ``def_``.

    If the attribute does not exist, raises :class:`AttributeError`.
    """
    suffixed = name + "_"
    # We can't call setattr() without asking first because if it doesn't exist it will be created,
    # which would be disastrous.
    if hasattr(obj, name):
        setattr(obj, name, value)
    elif hasattr(obj, suffixed):
        setattr(obj, suffixed, value)
    else:
        raise AttributeError(name)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import enum
import typing
import dataclasses
import pyuavcan


class FrameFormat(enum.IntEnum):
    BASE = 11
    EXTENDED = 29


@dataclasses.dataclass(frozen=True)
class DataFrame:
    format: FrameFormat
    identifier: int
    data: bytearray

    def __post_init__(self) -> None:
        assert isinstance(self.format, FrameFormat)
        if not (0 <= self.identifier < 2 ** int(self.format)):
            raise ValueError(f"Invalid CAN ID for format {self.format}: {self.identifier}")

        if len(self.data) not in _LENGTH_TO_DLC:
            raise ValueError(f"Unsupported data length: {len(self.data)}")

    @property
    def dlc(self) -> int:
        """Not to be confused with ``len(data)``."""
        return _LENGTH_TO_DLC[len(self.data)]  # The length is checked at the time of construction

    @staticmethod
    def convert_dlc_to_length(dlc: int) -> int:
        try:
            return _DLC_TO_LENGTH[dlc]
        except LookupError:
            raise ValueError(f"{dlc} is not a valid DLC") from None

    @staticmethod
    def get_required_padding(data_length: int) -> int:
        """
        Computes padding to nearest valid CAN FD frame size.

        >>> DataFrame.get_required_padding(6)
        0
        >>> DataFrame.get_required_padding(61)
        3
        """
        supremum = next(x for x in _DLC_TO_LENGTH if x >= data_length)  # pragma: no branch
        assert supremum >= data_length
        return supremum - data_length

    def __repr__(self) -> str:
        ide = {
            FrameFormat.EXTENDED: "0x%08x",
            FrameFormat.BASE: "0x%03x",
        }[self.format] % self.identifier
        return pyuavcan.util.repr_attributes(self, id=ide, data=self.data.hex())


@dataclasses.dataclass(frozen=True)
class Envelope:
    """
    The envelope models a singular input/output frame transaction.
    It is a media layer frame extended with IO-related metadata.
    """

    frame: DataFrame
    loopback: bool
    """Loopback request for outgoing frames; loopback indicator for received frames."""


_DLC_TO_LENGTH = [0, 1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 32, 48, 64]
_LENGTH_TO_DLC: typing.Dict[int, int] = dict(zip(*list(zip(*enumerate(_DLC_TO_LENGTH)))[::-1]))  # type: ignore
assert len(_LENGTH_TO_DLC) == 16 == len(_DLC_TO_LENGTH)
for item in _DLC_TO_LENGTH:
    assert _DLC_TO_LENGTH[_LENGTH_TO_DLC[item]] == item, "Invalid DLC tables"


def _unittest_can_media_frame() -> None:
    from pytest import raises

    for fmt in FrameFormat:
        with raises(ValueError):
            DataFrame(fmt, -1, bytearray())

        with raises(ValueError):
            DataFrame(fmt, 2 ** int(fmt), bytearray())

    with raises(ValueError):
        DataFrame(FrameFormat.EXTENDED, 123, bytearray(b"a" * 9))

    with raises(ValueError):
        DataFrame.convert_dlc_to_length(16)

    for sz in range(100):
        try:
            f = DataFrame(FrameFormat.EXTENDED, 123, bytearray(b"a" * sz))
        except ValueError:
            pass
        else:
            assert f.convert_dlc_to_length(f.dlc) == sz
